"""Contains the Playfield device class which represents the actual playfield in a pinball machine."""
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.ball_search import BallSearch
from mpf.core.delays import DelayManager
from mpf.devices.ball_device.incoming_balls_handler import IncomingBall


@DeviceMonitor("available_balls", "balls",
               num_balls_requested="balls_requested")
class Playfield(SystemWideDevice):

    """One playfield in a pinball machine."""

    config_section = 'playfields'
    collection = 'playfields'
    class_label = 'playfield'

    __slots__ = ["ball_search", "delay", "_balls", "available_balls", "num_balls_requested", "_incoming_balls",
                 "__dict__"]

    def __init__(self, machine, name):
        """Create the playfield."""
        super().__init__(machine, name)
        self.ball_search = BallSearch(self.machine, self)
        """An instance of :class:`mpf.core.ball_search.BallSearch` which
        handles ball search for this playfield."""

        self.delay = DelayManager(self.machine)
        """An instance of :class:`mpf.core.delays.DelayManager` which
        handles delays for this playfield."""

        self.machine.ball_devices[name] = self

        # Attributes
        self._balls = 0
        self.available_balls = 0
        self.num_balls_requested = 0

        self._incoming_balls = []

    async def _initialize(self):
        await super()._initialize()
        if 'default' in self.config['tags']:
            self.machine.playfield = self

        # Set up event handlers

        # Watch for balls added to the playfield
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                continue
            for target in device.config['eject_targets']:
                if target.name == self.name:
                    self.machine.events.add_handler(
                        event='balldevice_' + device.name +
                        '_ball_eject_success',
                        handler=self._source_device_eject_success)
                    self.machine.events.add_handler(
                        event='balldevice_{}_ball_lost'.format(device.name),
                        handler=self._source_device_ball_lost)
                    self.machine.events.add_handler(
                        event='balldevice_' + device.name +
                        '_ball_eject_failed',
                        handler=self._source_device_eject_failed)
                    self.machine.events.add_handler(
                        event='balldevice_' + device.name +
                        '_ejecting_ball',
                        handler=self._source_device_ejecting_ball)
                    break

        # Watch for balls removed from the playfield
        self.machine.events.add_handler('balldevice_captured_from_' + self.name,
                                        self._ball_removed_handler)

        # Watch for any switch hit which indicates a ball on the playfield
        self.machine.events.add_handler('sw_' + self.name + '_active',
                                        self._playfield_switch_hit)

        for device in self.machine.playfield_transfers.values():
            if device.config['eject_target'] == self:
                self.machine.events.add_handler(
                    event='balldevice_' + device.name +
                    '_ball_eject_success',
                    handler=self._source_device_eject_success)
                self.machine.events.add_handler(
                    event='balldevice_' + device.name +
                    '_ejecting_ball',
                    handler=self._source_device_ejecting_ball)

    async def expected_ball_received(self):
        """Handle an expected ball."""
        # We do nothing in that case

    async def unexpected_ball_received(self):
        """Handle an unexpected ball."""
        # We do nothing in that case

    @staticmethod
    async def wait_for_ready_to_receive(source):
        """Playfield is always ready to receive."""
        del source
        return True

    def add_missing_balls(self, balls):
        """Notify the playfield that it probably received a ball which went missing elsewhere."""
        # TODO: add incoming ball only and wait for confirm or timeout
        self.available_balls += balls
        self.balls += balls

    @property
    def balls(self):
        """Return the number of balls on the playfield."""
        return self._balls

    @balls.setter
    def balls(self, balls):

        prior_balls = self._balls
        ball_change = balls - prior_balls

        if ball_change:
            self.debug_log("Ball count change. Prior: %s, Current: %s, Change:"
                           " %s", prior_balls, balls, ball_change)

        self._balls = balls

        self.debug_log("New Ball Count: %s. (Prior count: %s)",
                       self._balls, prior_balls)

        if ball_change > 0:
            self.machine.events.post_relay('balldevice_' + self.name +
                                           '_ball_enter', new_balls=ball_change,
                                           unclaimed_balls=ball_change,
                                           device=self)
        # event docstring covered in base class

        if ball_change:
            self.machine.events.post(self.name + '_ball_count_change',
                                     balls=balls, change=ball_change)
        '''event: (name)_ball_count_change

        desc: The playfield with the name (name) has changed the number
        of balls that are live.

        args:
        balls: The current number of balls on the playfield.
        change: The change in balls from the last count.
        '''

        if balls <= 0:
            self.ball_search.disable()
        else:
            self.ball_search.enable()

    @classmethod
    def get_additional_ball_capacity(cls):
        """Return the number of ball which can be added.

        Used to find out how many more balls this device can hold. Since this
        is the playfield device, this method always returns 999.

        Returns: 999

        """
        return 999

    def add_ball(self, balls=1, source_device=None,
                 player_controlled=False) -> bool:
        """Add live ball(s) to the playfield.

        Args:
        ----
            balls: Integer of the number of balls you'd like to add.
            source_device: Optional ball device object you'd like to add the
                ball(s) from.
            player_controlled: Boolean which specifies whether this event is
                player controlled. (See not below for details)

        Returns True if it's able to process the add_ball() request, False if it
        cannot.

        The source_device arg is included to give you an options for specifying
        the source of the ball(s) to be added. This argument is optional, so if
        you don't supply them then MPF will use the default_source_device of
        this playfield.

        This method does *not* increase the game controller's count of the
        number of balls in play. So if you want to add balls (like in a
        multiball scenario), you need to call this method along with
        ``self.machine.game.add_balls_in_play()``.)

        MPF tracks the number of balls in play separately from the actual balls
        on the playfield because there are numerous situations where the two
        counts are not the same. For example, if a ball is in a VUK while some
        animation is playing, there are no balls on the playfield but still one
        ball in play, or if the player has a two-ball multiball and they shoot
        them both into locks, there are still two balls in play even though
        there are no balls on the playfield. The opposite can also be true,
        like when the player tilts then there are still balls on the playfield
        but no balls in play.

        Explanation of the player_controlled parameter:

        Set player_controlled to True to indicate that MPF should wait for the
        player to eject the ball from the source_device rather than firing a
        coil. The logic works like this:

        If the source_device does not have an eject_coil defined, then it's
        assumed that player_controlled is the only option. (e.g. this is a
        traditional plunger.) If the source_device does have an eject_coil
        defined, then there are two ways the eject could work. (1) there could
        be a "launch" button of some kind that's used to fire the eject coil,
        or (2) the device could be the auto/manual combo style where there's a
        mechanical plunger but also a coil which can eject the ball.

        If player_controlled is true and the device has an eject_coil, MPF will
        look for the player_controlled_eject_tag and eject the ball when a
        switch with that tag is activated.

        If there is no player_controlled_eject_tag, MPF assumes it's a manual
        plunger and will wait for the ball to disappear from the device based
        on the device's ball count decreasing.
        """
        if balls == 0:
            return False
        if balls < 0:
            raise AssertionError("Received request to add negative balls, which "
                                 "doesn't  make sense. Not adding any balls...")

        # Figure out which device we'll get a ball from
        if not source_device:
            source_device = self.config['default_source_device']

        if not source_device:
            self.raise_config_error("Missing a source device to request a ball. Did you define a default_source_device "
                                    "in your playfield?", 1)

        self.debug_log("Received request to add %s ball(s). Source device: %s."
                       " Player-controlled: %s", balls,
                       source_device.name, player_controlled)

        if player_controlled:
            for _ in range(balls):
                source_device.setup_player_controlled_eject(target=self)
        else:
            source_device.eject(balls=balls, target=self)

        return True

    def ball_arrived(self):
        """Confirm first ball in queue."""
        for incoming_ball in self._incoming_balls:
            # skip balls which cannot have possibly arrived
            if not incoming_ball.can_arrive:
                continue

            self.debug_log("Received ball from %s", incoming_ball.source)
            # confirm eject
            incoming_ball.ball_arrived()
            break

    def _mark_playfield_active(self):
        self.ball_arrived()
        self.machine.events.post_boolean(self.name + "_active")
        '''event: (name)_active
        desc: The playfield called (name) is now active, meaning there's
        at least one loose ball on it.
        '''

    def mark_playfield_active_from_device_action(self):
        """Mark playfield active because a device on the playfield detected activity."""
        self._playfield_switch_hit()

    def _playfield_switch_hit(self, **kwargs):
        """Playfield switch was hit.

        A switch tagged with '<this playfield name>_active' was just hit,
        indicating that there is at least one ball on the playfield.

        """
        if self.balls <= 0 or (kwargs.get('balls') and self.balls - kwargs['balls'] < 0):
            self._mark_playfield_active()

            if not self.num_balls_requested:
                self.debug_log("Playfield was activated with no balls expected.")
                self.machine.events.post('unexpected_ball_on_' + self.name)
                '''event: unexpected_ball_on_(name)
                desc: The playfield named (name) just had a switch hit,
                meaning a ball is on it, but that ball was not expected.
                '''

        self.ball_search.reset_timer()

    def _ball_removed_handler(self, balls, **kwargs):
        del kwargs
        # somebody got a ball from us so we obviously had one
        self.machine.events.post('sw_' + self.name + "_active",
                                 callback=self._ball_removed_handler2,
                                 balls=balls)
        '''event: sw_(name)_active
        desc: The playfield called (name) was active, though a ball
        was just removed from it.

        args:
        balls: The number of balls that were just removed from this playfield.
        '''

    def _ball_removed_handler2(self, balls):
        self.debug_log("%s ball(s) removed from the playfield", balls)
        self.balls -= balls
        self.available_balls -= balls
        for _ in range(balls):
            self.machine.ball_controller.add_captured_ball(self)

    def _source_device_ball_lost(self, target, **kwargs):
        del kwargs
        if target == self:
            self.available_balls -= 1

    def _source_device_ejecting_ball(self, balls, target, **kwargs):
        # A source device is attempting to eject a ball. We need to know if it's
        # headed to the playfield.
        del kwargs
        if target == self:
            self.debug_log("A source device is attempting to eject %s ball(s)"
                           " to the playfield.", balls)
            self.num_balls_requested += balls

    def _source_device_eject_failed(self, balls, target, **kwargs):
        # A source device failed to eject a ball. We need to know if it was
        # headed to the playfield.
        del kwargs
        if target == self:
            self.debug_log("A source device has failed to eject %s ball(s)"
                           " to the playfield.", balls)
            self.num_balls_requested -= balls

    def _source_device_eject_success(self, balls, target, **kwargs):
        del kwargs
        # A source device has just confirmed that it has successfully ejected a
        # ball. Note that we don't care what type of confirmation it used.
        # (Playfield switch hit, count of its ball switches, etc.)

        if target == self:
            self.debug_log("A source device has confirmed it's ejected %s "
                           "ball(s) to the playfield.", balls)
            self.balls += balls
            self.num_balls_requested -= balls

            if self.num_balls_requested < 0:
                raise AssertionError("num_balls_requested is smaller 0, which doesn't make sense. Quitting...")

    @classmethod
    def is_playfield(cls):
        """Return true since it is a playfield."""
        return True

    def add_incoming_ball(self, incoming_ball: IncomingBall):
        """Track an incoming ball."""
        self._incoming_balls.append(incoming_ball)

    def remove_incoming_ball(self, incoming_ball: IncomingBall):
        """Stop tracking an incoming ball."""
        try:
            self._incoming_balls.remove(incoming_ball)
        except ValueError as e:
            if self.unit_test:
                # re-raise this in tests
                raise e
            self.warning_log("Double remove of incoming ball. This is likely a bug! "
                             "Please report in the forum if you can reproduce it.")

    @event_handler(1)
    def event_ball_search_disable(self, **kwargs):
        """Disable ball search for this playfield.

        If the ball search timer is running, it will stop and disable it. If
        an actual ball search process is running, it will stop.
        """
        del kwargs
        self.ball_search.disable()

    @event_handler(2)
    def event_ball_search_enable(self, **kwargs):
        """Enable ball search for this playfield.

        Note this does not start the ball search process, rather, it starts the
        timer running.
        """
        del kwargs
        self.ball_search.enable()

    @event_handler(3)
    def event_ball_search_block(self, **kwargs):
        """Block ball search for this playfield.

        Blocking will disable ball search if it's enabled or running, and will
        prevent ball search from enabling if it's disabled until
        ball_search_resume() is called.
        """
        del kwargs
        self.ball_search.block()

    @event_handler(4)
    def event_ball_search_unblock(self, **kwargs):
        """Unblock ball search for this playfield.

        This will check to see if there are balls on the playfield, and if so,
        enable ball search.
        """
        del kwargs
        self.ball_search.unblock()



"""RGB DMD on the RPi.

Contains support for the Raspi low-level RGB LED tile driver of hzeller
(https://github.com/hzeller/rpi-rgb-led-matrix).
"""
import atexit
from PIL import Image

from mpf.core.platform import RgbDmdPlatform
from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface

# the hzeller library (external dependency)
# pylint: disable-msg=ungrouped-imports
try:
    from rgbmatrix import RGBMatrix, RGBMatrixOptions
except ImportError:
    RGBMatrix = None
    RGBMatrixOptions = None


class RpiDmdPlatform(RgbDmdPlatform):

    """Raspberry Pi GPIO RGB DMD."""

    __slots__ = ["_dmd", "config"]

    def __init__(self, machine):
        """Initialise RGB DMD."""
        super().__init__(machine)
        if not RGBMatrix:
            raise AssertionError("Please install the rgbmatrix or RGB Matrix library to use a RGB Matrix on the RPi")

        self.features['tickless'] = True
        self._dmd = None
        self.config = None
        atexit.register(self.stop)

    async def initialize(self):
        """Initialise platform."""
        self.config = self.machine.config_validator.validate_config(
            config_spec='rpi_dmd',
            source=self.machine.config.get('rpi_dmd', {})
        )

    def stop(self):
        """Stop platform."""
        if self._dmd:
            self._dmd.stop()

    def __repr__(self):
        """Return string representation."""
        return '<Platform.RpiDmd>'

    def configure_rgb_dmd(self, name: str):
        """Configure rgb dmd."""
        if not self._dmd:
            self._dmd = RpiRgbDmdDevice(self.config)
        return self._dmd


# noinspection PyCallingNonCallable
class RpiRgbDmdDevice(DmdPlatformInterface):

    """A RpiRgbDmd device."""

    def __init__(self, config):
        """Initialise RpiRgbDmd device."""
        self.config = config
        xs = config["cols"]
        ys = config["rows"]
        self.img = Image.frombytes("RGB", (xs, ys), b'\x11' * xs * ys * 3)
        self.rgb_opts = RGBMatrixOptions()
        # Rudeboy way of setting the RGBMatrixOptions
        for k, v in config.items():
            try:
                setattr(self.rgb_opts, k, v)
            except Exception:   # pylint: disable-msg=broad-except
                raise AssertionError("RpiRgbDmdDevice: couldn't set", k, v)
        self.matrix = RGBMatrix(options=self.rgb_opts)
        self.matrix.SetImage(self.img)

    def update(self, data):
        """Update DMD data."""
        self.img.frombytes(data)
        self.matrix.SetImage(self.img)

    def set_brightness(self, brightness: float):
        """Set brightness.

        Range is [0.0 ... 1.0].
        """
        self.matrix.brightness = brightness * 100

    def stop(self):
        """Stop device."""
        self.matrix.Clear()



from mpf.tests.MpfGameTestCase import MpfGameTestCase
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import test_config


class TestBallSearch(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_search/'

    def get_platform(self):
        return 'smart_virtual'

    def test_ball_search_does_not_start_when_disabled(self):
        self.machine.playfields["playfield"].config['enable_ball_search'] = False

        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.advance_time_and_run(30)
        self.assertFalse(self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertFalse(self.machine.ball_devices['playfield'].ball_search.started)

        self.machine.playfields["playfield"].config['enable_ball_search'] = None
        self.machine.config['mpf']['default_ball_search'] = True
        self.machine.ball_devices['playfield'].ball_search.enable()
        self.assertTrue(self.machine.ball_devices['playfield'].ball_search.enabled)
        self.machine.ball_devices['playfield'].ball_search.disable()
        self.assertFalse(self.machine.ball_devices['playfield'].ball_search.enabled)

        self.machine.playfields["playfield"].config['enable_ball_search'] = None
        self.machine.config['mpf']['default_ball_search'] = False
        self.machine.ball_devices['playfield'].ball_search.enable()
        self.assertFalse(self.machine.ball_devices['playfield'].ball_search.enabled)

    def test_game_with_no_switches(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        # wait for eject_timeout of launcher
        self.advance_time_and_run(6)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)

        # vuk should reset the timer
        self.machine.switch_controller.process_switch("s_vuk", 1)
        # wait for eject_timeout of vuk
        self.advance_time_and_run(3)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.advance_time_and_run(15)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.advance_time_and_run(5)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # ball drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

    @test_config("config_with_balls.yaml")
    def test_game_with_switches(self):
        self.assertNumBallsKnown(2)

        # motor moves to reset position
        self.assertEqual("enabled", self.machine.digital_outputs["c_motor_run"].hw_driver.state)

        self.machine.switch_controller.process_switch("s_position_down", 1)
        self.advance_time_and_run(1)

        # and stops
        self.assertEqual("disabled", self.machine.digital_outputs["c_motor_run"].hw_driver.state)

        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(2)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)

        # ball entering and being ejected from vuk should reset search timer
        self.advance_time_and_run(15)
        self.machine.switch_controller.process_switch("s_vuk", 1)
        self.advance_time_and_run(15)

        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # pf switch should reset the timer
        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(15)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(15)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # advance past 20s since last hit and search should be running
        self.advance_time_and_run(6)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # hit a pf switch and the ball search should stop
        self.hit_switch_and_run('s_playfield', 1)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # advance 20s and search should be running again
        self.advance_time_and_run(20)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # ball entering vuk should stop the search
        self.hit_switch_and_run('s_vuk', 2)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # test servo reset during ball search
        self.machine.servos["servo1"].go_to_position(0.7)
        self.advance_time_and_run(21)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        self.assertEqual(0.0, self.machine.servos["servo1"].hw_servo.current_position)
        self.advance_time_and_run(5)
        self.assertEqual(1.0, self.machine.servos["servo1"].hw_servo.current_position)
        self.advance_time_and_run(5)
        self.assertEqual(0.0, self.machine.servos["servo1"].hw_servo.current_position)

        # test motor. should be running
        self.assertEqual("enabled", self.machine.digital_outputs["c_motor_run"].hw_driver.state)

        # it leaves the down position
        self.machine.switch_controller.process_switch("s_position_down", 0)
        self.advance_time_and_run(1)

        # moves by the original position
        self.machine.switch_controller.process_switch("s_position_down", 0)
        self.advance_time_and_run(1)

        # it leaves the down position again
        self.machine.switch_controller.process_switch("s_position_down", 0)
        self.advance_time_and_run(0)

        # ball drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # servo should return to its position
        self.assertEqual(0.7, self.machine.servos["servo1"].hw_servo.current_position)

        # motor keeps running
        self.assertEqual("enabled", self.machine.digital_outputs["c_motor_run"].hw_driver.state)

        # until it reaches its original position
        self.machine.switch_controller.process_switch("s_position_down", 1)
        self.advance_time_and_run(1)
        self.assertEqual("disabled", self.machine.digital_outputs["c_motor_run"].hw_driver.state)

    def test_ball_search_iterations_and_give_up_with_new_ball(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(2)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(.1)

        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)

        # push down target 3 and 4
        self.machine.switch_controller.process_switch("s_drop_target3", 1)
        self.machine.switch_controller.process_switch("s_drop_target4", 1)

        self.advance_time_and_run(10.1)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.machine.ball_devices['playfield'].add_ball = MagicMock()

        for i in range(1, 11):
                # this will break smart_virtual
                self.machine.coils['eject_coil1'].pulse = MagicMock()
                self.machine.coils['eject_coil2'].pulse = MagicMock()
                self.machine.coils['eject_coil3'].pulse = MagicMock()
                self.machine.coils['hold_coil'].pulse = MagicMock()
                self.machine.coils['drop_target_reset1'].pulse = MagicMock()
                self.machine.coils['drop_target_reset2'].pulse = MagicMock()
                self.machine.coils['drop_target_reset3'].pulse = MagicMock()
                self.machine.coils['drop_target_reset4'].pulse = MagicMock()
                self.machine.coils['drop_target_knockdown2'].pulse = MagicMock()
                self.machine.coils['drop_target_knockdown4'].pulse = MagicMock()
                self.machine.coils['flipper_coil'].enable = MagicMock()
                self.machine.coils['diverter_coil'].enable = MagicMock()
                self.machine.coils['autofire_coil'].pulse = MagicMock()
                self.advance_time_and_run(10)
                if i <= 3:
                    self.assertEqual(1, self.machine.ball_devices['playfield'].ball_search.phase)
                elif i <= 6:
                    self.assertEqual(2, self.machine.ball_devices['playfield'].ball_search.phase)
                else:
                    self.assertEqual(3, self.machine.ball_devices['playfield'].ball_search.phase)

                self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

                assert not self.machine.coils['eject_coil1'].pulse.called
                if 3 < i <= 6:
                    self.machine.coils['eject_coil2'].pulse.assert_called_with(5)
                else:
                    self.machine.coils['eject_coil2'].pulse.assert_called_with()
                assert not self.machine.coils['eject_coil3'].pulse.called
                assert not self.machine.coils['hold_coil'].pulse.called
                assert not self.machine.coils['drop_target_reset1'].pulse.called
                assert not self.machine.coils['drop_target_reset2'].pulse.called
                assert not self.machine.coils['drop_target_knockdown2'].pulse.called

                self.advance_time_and_run(.25)

                assert not self.machine.coils['eject_coil1'].pulse.called
                self.machine.coils['eject_coil3'].pulse.assert_called_with()
                assert not self.machine.coils['hold_coil'].pulse.called
                assert not self.machine.coils['drop_target_reset1'].pulse.called
                assert not self.machine.coils['drop_target_reset2'].pulse.called
                assert not self.machine.coils['drop_target_knockdown2'].pulse.called

                self.advance_time_and_run(.25)

                assert not self.machine.coils['eject_coil1'].pulse.called
                self.machine.coils['hold_coil'].pulse.assert_called_with()
                assert not self.machine.coils['drop_target_reset1'].pulse.called
                assert not self.machine.coils['drop_target_reset2'].pulse.called
                assert not self.machine.coils['drop_target_knockdown2'].pulse.called

                self.advance_time_and_run(.25)

                assert not self.machine.coils['eject_coil1'].pulse.called
                self.machine.coils['drop_target_reset1'].pulse.assert_called_with()
                assert not self.machine.coils['drop_target_reset2'].pulse.called
                assert not self.machine.coils['drop_target_knockdown2'].pulse.called

                self.advance_time_and_run(.25)

                assert not self.machine.coils['eject_coil1'].pulse.called
                if i <= 3:
                    self.machine.coils['drop_target_reset2'].pulse.assert_called_with()
                    assert not self.machine.coils['drop_target_knockdown2'].pulse.called
                else:
                    self.machine.coils['drop_target_reset2'].pulse.assert_called_with()
                    self.machine.coils['drop_target_knockdown2'].pulse.assert_called_with()

                assert not self.machine.coils['drop_target_reset3'].pulse.called
                assert not self.machine.coils['drop_target_reset4'].pulse.called
                assert not self.machine.coils['drop_target_knockdown4'].pulse.called

                if i > 6:
                    self.advance_time_and_run(.25)
                    self.machine.coils['drop_target_reset3'].pulse.assert_called_with()

                self.advance_time_and_run(.25)
                if i <= 3:
                    self.machine.coils['drop_target_knockdown4'].pulse.assert_called_with()
                    assert not self.machine.coils['drop_target_reset4'].pulse.called
                else:
                    self.machine.coils['drop_target_reset4'].pulse.assert_called_with()
                    self.machine.coils['drop_target_knockdown4'].pulse.assert_called_with()

                assert not self.machine.ball_devices['playfield'].add_ball.called

                self.advance_time_and_run(.25)
                assert self.machine.coils['diverter_coil'].enable.called
                assert not self.machine.coils['flipper_coil'].enable.called
                assert not self.machine.coils['autofire_coil'].pulse.called

                self.advance_time_and_run(.25)
                assert self.machine.coils['flipper_coil'].enable.called
                assert not self.machine.coils['autofire_coil'].pulse.called

                self.advance_time_and_run(.25)
                assert self.machine.coils['autofire_coil'].pulse.called

        self.advance_time_and_run(10)

        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.machine.ball_devices['playfield'].add_ball.assert_called_with()

    def test_give_up_with_game_end_in_game(self):
        self.machine.ball_devices['playfield'].config['ball_search_failed_action'] = 'end_game'

        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.advance_time_and_run(30)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # wait for ball search to fail
        self.advance_time_and_run(300)
        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(None, self.machine.game)

    def test_give_up_with_game_end_outside_game(self):
        self.machine.ball_devices['playfield'].config['ball_search_failed_action'] = 'end_game'

        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.assertEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        # add a random ball to the pf via vuk
        self.machine.switch_controller.process_switch("s_vuk", 1)
        self.advance_time_and_run(30)

        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # try to start a game (should not work)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

        # wait for ball search to fail
        self.advance_time_and_run(300)
        self.assertEqual(None, self.machine.game)
        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(3, self.machine.ball_controller.num_balls_known)

        # try to start a game (should work again)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)

    def test_give_up_with_end_ball(self):
        self.machine.playfields["playfield"].config['ball_search_failed_action'] = "end_ball"

        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.start_game()
        self.advance_time_and_run(10)
        self.assertBallNumber(1)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.advance_time_and_run(30)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # wait for ball search to fail
        self.advance_time_and_run(100)
        self.assertBallNumber(2)
        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)

    def test_give_up_with_no_more_balls(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.machine.coils['eject_coil1'].pulse = MagicMock()
        self.machine.coils['eject_coil2'].pulse = MagicMock()
        self.machine.coils['eject_coil3'].pulse = MagicMock()
        self.machine.coils['hold_coil'].pulse = MagicMock()

        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.advance_time_and_run(30)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # wait for ball search to fail
        self.advance_time_and_run(150)
        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(0, self.machine.ball_controller.num_balls_known)
        self.assertEqual(None, self.machine.game)

    @test_config("missing_initial.yaml")
    def test_missing_initial(self):
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self.machine.playfield.available_balls)

        self.assertTrue(self.machine.playfield.ball_search.started)
        self.advance_time_and_run(20)
        self.hit_switch_and_run("s_ball_switch2", 1)

        self.assertFalse(self.machine.playfield.ball_search.started)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self.machine.playfield.available_balls)

    @test_config("missing_initial.yaml")
    def test_missing_initial2(self):
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self.machine.playfield.available_balls)

        self.assertTrue(self.machine.playfield.ball_search.started)
        self.advance_time_and_run(20)

        # game does not start because of ball search
        self.assertEqual(None, self.machine.game)
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

        # cancel ball search
        self.post_event("cancel_ball_search")
        self.advance_time_and_run(1)

        # game should start
        self.assertEqual(None, self.machine.game)
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)

    def test_ball_search_cancel_running(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)
        self.machine.diverters["diverter1"].activate()
        self.assertEqual("enabled", self.machine.coils["diverter_coil"].hw_driver.state)

        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.advance_time_and_run(21)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

        # diverter disables during ball search
        self.assertEqual("disabled", self.machine.coils["diverter_coil"].hw_driver.state)

        self.post_event('flipper_cradle', 1)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        # diverter should restore state
        self.assertEqual("enabled", self.machine.coils["diverter_coil"].hw_driver.state)

    def test_prevent_ball_search_with_flipper_cradle(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)

        self.assertEqual(None, self.machine.game)
        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)

        self.post_event('flipper_cradle', 1)
        self.advance_time_and_run(21)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.post_event('flipper_cradle_release', 1)
        self.advance_time_and_run(21)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

    @test_config("mechanical_eject.yaml")
    def test_mechanical_eject(self):
        self.mock_event("ball_search_started")
        self.hit_switch_and_run("s_ball_switch1", 0)
        self.hit_switch_and_run("s_ball_switch2", 0)
        self.hit_switch_and_run("s_ball_switch3", 1)

        self.start_game()
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(1, self.machine.ball_devices["test_launcher"].balls)
        self.advance_time_and_run(40)
        self.assertEqual(1, self.machine.ball_devices["test_launcher"].balls)
        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.release_switch_and_run("s_ball_switch_launcher", 15)
        self.assertEqual(0, self.machine.ball_devices["test_launcher"].balls)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        # ball returns to plunger
        self.hit_switch_and_run("s_ball_switch_launcher", 30)
        self.assertEqual(1, self.machine.ball_devices["test_launcher"].balls)
        self.assertEqual(0, self.machine.ball_devices['playfield'].balls)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.assertEventNotCalled("ball_search_started")

        # just to make sure we got the right event
        self.release_switch_and_run("s_ball_switch_launcher", 40)
        self.assertEventCalled("ball_search_started")
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

    @test_config("no_eject.yaml")
    def test_ball_search_no_eject(self):
        self.machine.ball_controller.num_balls_known = 0
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.machine.switch_controller.process_switch("s_ball_switch3", 1)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_start", 1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(2)
        self.assertNotEqual(None, self.machine.game)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(.1)

        self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.enabled)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)
        self.assertEqual(1, self.machine.ball_devices['playfield'].balls)

        self.advance_time_and_run(10.1)
        self.assertEqual(False, self.machine.ball_devices['playfield'].ball_search.started)

        self.machine.ball_devices['playfield'].add_ball = MagicMock()

        phases = {}
        for i in range(1, 11):
                # this will break smart_virtual
                self.machine.coils['eject_coil1'].pulse = MagicMock()
                self.machine.coils['eject_coil2'].pulse = MagicMock()
                self.machine.coils['eject_coil3'].pulse = MagicMock()
                self.machine.coils['hold_coil'].pulse = MagicMock()
                self.machine.coils['drop_target_reset1'].pulse = MagicMock()
                self.machine.coils['drop_target_reset2'].pulse = MagicMock()
                self.machine.coils['drop_target_reset3'].pulse = MagicMock()
                self.machine.coils['drop_target_reset4'].pulse = MagicMock()
                self.machine.coils['drop_target_knockdown2'].pulse = MagicMock()
                self.machine.coils['drop_target_knockdown4'].pulse = MagicMock()
                self.machine.coils['flipper_coil'].enable = MagicMock()
                self.machine.coils['diverter_coil'].enable = MagicMock()
                self.machine.coils['autofire_coil'].pulse = MagicMock()
                self.advance_time_and_run(10)
                # Record that the current phase was covered
                phase = self.machine.ball_devices['playfield'].ball_search.phase
                phases[phase] = True

                self.assertEqual(True, self.machine.ball_devices['playfield'].ball_search.started)

                # Make sure the right interaction ocurrs with the trough
                assert not self.machine.coils['eject_coil1'].pulse.called
                # Check to make sure the right interaction with the test
                # launcher occurred during the phase of ball search
                self.machine.coils['eject_coil2'].pulse.assert_not_called()

                self.advance_time_and_run(2)
                if i > 6:
                    self.advance_time_and_run(.25)

        self.advance_time_and_run(10)
        # make sure that all relevant phases have been run through
        for phase in range(1, 4):
            assert phase in phases



"""Score reel controller."""
import logging

import asyncio
from functools import partial


class ScoreReelController:

    """The overall controller that is in charge of and manages the score reels in a pinball machine.

    The main thing this controller does is keep track of how many
    ScoreReelGroups there are in the machine and how many players there are,
    as well as maps the current player to the proper score reel.

    This controller is also responsible for working around broken
    ScoreReelGroups and "stacking" and switching out players when there are
    multiple players per ScoreReelGroup.

    Known limitations of this module:
        * Assumes all score reels include a zero value.
        * Assumes all score reels count up or down by one.
        * Assumes all score reels map their displayed value to their stored
          value in a 1:1 way. (i.e. value[0] displays 0, value[5] displays 5,
          etc.
        * Currently this module only supports "incrementing" reels (i.e.
          counting up). Decrementing support will be added in the future.
    """

    config_name = "score_reel_controller"

    def __init__(self, machine):
        """Initialise score reel controller."""
        self.machine = machine
        self.log = logging.getLogger("ScoreReelController")
        self.log.debug("Loading the ScoreReelController")

        self.active_scorereelgroup = None
        """Pointer to the active ScoreReelGroup for the current player.
        """
        self.player_to_scorereel_map = {}
        """This is a dict of ScoreReelGroup objects which corresponds to player
        number. The first element [1] in this dict is the first player (which
        is player number [1], the next one is the next player, etc.
        """

        self.active_reel_player_map = {}
        """This dict maps reels to players. Every reel can only have one active player but multiple players can
        share a reel.
        """
        # switch the active score reel group and reset it (if needed)
        self.machine.events.add_handler('player_turn_started',
                                        self._rotate_player)

        # receive notification of score changes
        self.machine.events.add_handler('player_score', self._score_change)

        # receives notifications of game starts to reset the reels
        self.machine.events.add_handler('game_starting', self._game_starting)

        # receives notifications of game ends to reset the reels
        self.machine.events.add_handler('game_ending', self._game_ending)

        # Need to hook this in case reels aren't done when ball ends
        self.machine.events.add_handler('ball_ending', self._ball_ending, 900)

    def _rotate_player(self, **kwargs):
        """Start a new player's turn.

        The main purpose of this method is to map the current player to their
        ScoreReelGroup in the backbox. It will do this by comparing length of
        the list which holds those mappings (`player_to_scorereel_map`) to
        the length of the list of players. If the player list is longer that
        means we don't have a ScoreReelGroup for that player.

        In that case it will check the tags of the ScoreReelGroups to see if
        one of them is tagged with playerX which corresponds to this player.
        If not then it will pick the next free one. If there are none free,
        then it will "double up" that player on an existing one which means
        the same Score Reels will be used for both players, and they will
        reset themselves automatically between players.
        """
        del kwargs
        # unlight active score reel group
        if self.active_scorereelgroup:
            self.active_scorereelgroup.unlight()

        self.active_scorereelgroup = self.player_to_scorereel_map[self.machine.game.player.number]

        self.active_reel_player_map[self.active_scorereelgroup] = self.machine.game.player.number

        self.log.debug("Mapping Player %s to ScoreReelGroup '%s'",
                       self.machine.game.player.number,
                       self.active_scorereelgroup.name)

        # Make sure this score reel group is showing the right score
        self.log.debug("Current player's score: %s",
                       self.machine.game.player.score)
        self.active_scorereelgroup.set_value(self.machine.game.player.score)

        self.active_scorereelgroup.light()

    def _score_change(self, value, change, player_num, **kwargs):
        """Handle score changes and add the score increase to the current active ScoreReelGroup.

        This method is the handler for the score change event, so it's called
        automatically.

        Args:
        ----
            value: Integer value of the new score. This parameter is ignored,
                and included only because the score change event passes it.
            change: Change compared to the previous score-
            player_num: Player number of the player who's score changed.
        """
        del kwargs
        del change
        # get score reel group for player
        score_reel_group = self.player_to_scorereel_map[player_num]

        # check if it is currently dedicated to that player
        if score_reel_group and self.active_reel_player_map[score_reel_group] == player_num:
            # set value
            score_reel_group.set_value(value=value)

    def _game_starting(self, queue, **kwargs):
        """Reset the score reels when a new game starts.

        This is a queue event so it doesn't allow the game start to continue
        until it's done.

        Args:
        ----
            queue: A reference to the queue object for the game starting event.
        """
        del kwargs
        # tell the game_starting event queue that we have stuff to do
        queue.wait()

        # calculate a player <-> reel mapping
        for player_num in range(1, self.machine.game.max_players + 1):
            reel = self.machine.score_reel_groups.items_tagged("player{}".format(player_num))
            if reel:
                self.player_to_scorereel_map[player_num] = reel[0]
                if reel[0] not in self.active_reel_player_map:
                    self.active_reel_player_map[reel[0]] = player_num
            else:
                self.log.warning('Did not find a score reel for player %s. Did you tag a reel with "player%s"? '
                                 'Will reuse player1',
                                 player_num, player_num)
                reel1 = self.machine.score_reel_groups.items_tagged("player1")
                self.player_to_scorereel_map[player_num] = reel1[0]
                if not reel1:
                    raise AssertionError('Need a score reel group tagged "player1"')

        futures = []
        for score_reel_group in self.machine.score_reel_groups.values():
            score_reel_group.set_value(0)
            futures.append(score_reel_group.wait_for_ready())

        future = asyncio.wait(iter(futures))
        future = asyncio.ensure_future(future)
        future.add_done_callback(partial(self._reels_ready, queue=queue))

    @staticmethod
    def _reels_ready(future, queue):
        """Unblock queue since all reels are ready."""
        del future
        queue.clear()

    def _game_ending(self, **kwargs):
        """Reset controller."""
        del kwargs
        if self.active_scorereelgroup:
            self.active_scorereelgroup.unlight()
        self.active_scorereelgroup = None
        self.player_to_scorereel_map = {}

    def _ball_ending(self, queue=None, **kwargs):
        del kwargs
        # We need to hook the ball_ending event in case the ball ends while the
        # score reel is still catching up.
        queue.wait()

        future = asyncio.ensure_future(self.active_scorereelgroup.wait_for_ready())
        future.add_done_callback(partial(self._reels_ready, queue=queue))



"""RPC Interface for BCP clients."""
from copy import deepcopy

from mpf.core.rgb_color import ColorException

from mpf.core.events import PostedEvent
from mpf.core.player import Player
from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController
from mpf.core.switch_controller import MonitoredSwitchChange
from mpf.exceptions.driver_limits_error import DriverLimitsError


class BcpInterface(MpfController):

    """Implements the BCP interface which can be used by all clients.

    Args:
    ----
        machine: A reference to the main MPF machine object.

    The following BCP commands are currently implemented:
        error
        get
        hello?version=xxx&controller_name=xxx&controller_version=xxx
        mode_start?name=xxx&priority=xxx
        mode_stop?name=xxx
        player_added?player_num=x
        player_variable?name=x&value=x&prev_value=x&change=x&player_num=x
        set
        shot?name=x
        switch?name=x&state=x
        timer
        trigger?name=xxx

    """

    config_name = "bcp_interface"

    __slots__ = ["configured", "config", "_client_reset_queue", "_client_reset_complete_status", "bcp_receive_commands",
                 "_shows"]

    def __init__(self, machine):
        """Initialise BCP."""
        super().__init__(machine)

        if 'bcp' not in machine.config or not machine.config['bcp']:
            self.configured = False
            return

        self.configured = True

        self.config = machine.config['bcp']

        self._client_reset_queue = None
        self._client_reset_complete_status = {}

        self.bcp_receive_commands = dict(
            reset_complete=self._bcp_receive_reset_complete,
            error=self._bcp_receive_error,
            switch=self._bcp_receive_switch,
            trigger=self._bcp_receive_trigger,
            register_trigger=self._bcp_receive_register_trigger,
            evaluate_placeholder=self._evaluate_placeholder,
            remove_trigger=self._bcp_receive_deregister_trigger,
            monitor_start=self._bcp_receive_monitor_start,
            monitor_stop=self._bcp_receive_monitor_stop,
            set_machine_var=self._bcp_receive_set_machine_var,
            service=self._service,
        )
        self._shows = {}

        self.machine.events.add_handler('machine_reset_phase_1', self.bcp_reset)

    def __repr__(self):
        """Return string representation."""
        return '<BCP Interface>'

    def register_command_callback(self, cmd, callback):
        """Register a BCP command."""
        if not self.configured:
            return
        self.bcp_receive_commands[cmd] = callback

    def add_registered_trigger_event_for_client(self, client, event):
        """Add trigger for event."""
        # register handler if first transport
        if not self.machine.bcp.transport.get_transports_for_handler(event):
            self.machine.events.add_handler(event=event,
                                            handler=self.bcp_trigger,
                                            name=event)
        # register transport
        self.machine.bcp.transport.add_handler_to_transport(event, client)

    def remove_registered_trigger_event_for_client(self, client, event):
        """Remove trigger for event."""
        # unregister transport
        self.machine.bcp.transport.remove_transport_from_handle(event, client)

        # if not transports remain, remove handler
        if not self.machine.bcp.transport.get_transports_for_handler(event):
            self.machine.events.remove_handler_by_event(event=event, handler=self.bcp_trigger)

    async def _bcp_receive_set_machine_var(self, client, name, value):
        """Set machine var via bcp."""
        del client
        self.machine.variables.set_machine_var(name, value)
        # document variables injected by MC
        '''machine_var: mc_version

        desc: Version of MC. This is set after MC got connected.
        '''
        '''machine_var: mc_extended_version

        desc: Extended version of MC. This is set after MC got connected. Contains BCP and show version numbers.
        '''

    async def _service_stop(self, client):
        for show in self._shows.values():
            show.stop()
        for light in self.machine.lights.values():
            light.remove_from_stack_by_key("service")
        self._shows = {}
        await self.machine.service.stop_service()
        self.machine.bcp.transport.send_to_client(client, "service_stop")

    async def _service(self, client, subcommand, **kwargs):
        """Run service command."""
        if subcommand == "start":
            self.machine.service.start_service()
        elif subcommand == "stop":
            await self._service_stop(client)
        elif subcommand == "list_switches":
            self.machine.bcp.transport.send_to_client(client, "list_switches",
                                                      switches=[(s[0], str(s[1].hw_switch.number), s[1].name,
                                                                 s[1].state)
                                                                for s in self.machine.service.get_switch_map()])
        elif subcommand == "list_coils":
            self.machine.bcp.transport.send_to_client(client, "list_coils",
                                                      coils=[(s[0], str(s[1].hw_driver.number), s[1].name) for s in
                                                             self.machine.service.get_coil_map()])
        elif subcommand == "list_lights":
            self.machine.bcp.transport.send_to_client(client, "list_lights",
                                                      lights=[(s[0], s[1].get_hw_numbers(), s[1].name, s[1].get_color())
                                                              for s in self.machine.service.get_light_map()])
        elif subcommand == "list_shows":
            self.machine.bcp.transport.send_to_client(client, "list_shows",
                                                      shows=[(s.name, sorted(s.tokens))
                                                             for s in sorted(self.machine.shows.values(),
                                                                             key=lambda x: x.name)])
        elif subcommand == "monitor_switches":
            pass
        elif subcommand == "coil_pulse":
            self._coil_pulse(client, kwargs.get("coil"), kwargs.get("pulse_ms"), kwargs.get("pulse_power"))
        elif subcommand == "coil_enable":
            self._coil_enable(client, kwargs.get("coil"), kwargs.get("pulse_ms"), kwargs.get("pulse_power"),
                              kwargs.get("hold_power"))
        elif subcommand == "coil_disable":
            self._coil_disable(client, kwargs.get("coil"))
        elif subcommand == "show_play":
            self._show_play(client, kwargs.get("show"), kwargs.get("token"))
        elif subcommand == "show_stop":
            self._show_stop(client, kwargs.get("show"))
        elif subcommand == "light_color":
            self._light_color(client, kwargs.get("light"), kwargs.get("color"))

    def _show_play(self, client, show_name, token):
        try:
            show = self.machine.shows[show_name]
        except KeyError:
            self.machine.bcp.transport.send_to_client(client, "show_play", error="Show not found")
            return
        if show_name in self._shows:
            self._shows[show_name].stop()
        try:
            self._shows[show_name] = show.play(show_tokens=token, priority=100000)
        except (ValueError, AssertionError) as e:
            self.machine.bcp.transport.send_to_client(client, "show_play", error="Show error: {}".format(e))
            return
        self.machine.bcp.transport.send_to_client(client, "show_play", error=False)

    def _show_stop(self, client, show_name):
        if show_name in self._shows:
            self._shows[show_name].stop()
            del self._shows[show_name]
            self.machine.bcp.transport.send_to_client(client, "show_stop", error=False)
        else:
            self.machine.bcp.transport.send_to_client(client, "show_stop", error="Show not playing")

    def _coil_pulse(self, client, coil_name, pulse_ms, pulse_power):
        try:
            coil = self.machine.coils[coil_name]
        except KeyError:
            self.machine.bcp.transport.send_to_client(client, "coil_pulse", error="Coil not found")
            return
        if pulse_ms:
            pulse_ms = int(pulse_ms)
        if pulse_power:
            pulse_power = float(pulse_power)
        coil.pulse(pulse_ms=pulse_ms, pulse_power=pulse_power)
        self.machine.bcp.transport.send_to_client(client, "coil_pulse", error=False)

    def _coil_disable(self, client, coil_name):
        try:
            coil = self.machine.coils[coil_name]
        except KeyError:
            self.machine.bcp.transport.send_to_client(client, "coil_disable", error="Coil not found")
            return
        coil.disable()
        self.machine.bcp.transport.send_to_client(client, "coil_disable", error=False)

    # pylint: disable-msg=too-many-arguments
    def _coil_enable(self, client, coil_name, pulse_ms, pulse_power, hold_power):
        try:
            coil = self.machine.coils[coil_name]
        except KeyError:
            self.machine.bcp.transport.send_to_client(client, "coil_enable", error="Coil not found")
            return
        if pulse_ms:
            pulse_ms = int(pulse_ms)
        if pulse_power:
            pulse_power = float(pulse_power)
        if hold_power:
            hold_power = float(hold_power)
        try:
            coil.enable(pulse_ms=pulse_ms, pulse_power=pulse_power, hold_power=hold_power)
        except DriverLimitsError as e:
            self.machine.bcp.transport.send_to_client(client, "coil_enable", error=str(e))
            return

        self.machine.bcp.transport.send_to_client(client, "coil_enable", error=False)

    def _light_color(self, client, light_name, color_name):
        try:
            light = self.machine.lights[light_name]
        except KeyError:
            self.machine.bcp.transport.send_to_client(client, "light_color", error="Light not found")
            return
        try:
            light.color(color_name, key="service")
        except (DriverLimitsError, ColorException) as e:
            self.machine.bcp.transport.send_to_client(client, "light_color", error=str(e))
            return

        self.machine.bcp.transport.send_to_client(client, "light_color", error=False)

    async def _bcp_receive_monitor_start(self, client, category):
        """Start monitoring the specified category."""
        category = str.lower(category)

        if category == "events":
            self._monitor_events(client)
        elif category == "devices":
            self._monitor_devices(client)
        elif category == "drivers":
            self._monitor_drivers(client)
        elif category == "switches":
            self._monitor_switches(client)
        elif category == "machine_vars":
            self._monitor_machine_vars(client)
        elif category == "player_vars":
            self._monitor_player_vars(client)
        elif category == "modes":
            self._monitor_modes(client)
        elif category == "core_events":
            self._monitor_core_events(client)
        elif category == "status_request":
            self._monitor_status_request(client)
        else:
            self.machine.bcp.transport.send_to_client(client,
                                                      "error",
                                                      cmd="monitor_start?category={}".format(category),
                                                      error="Invalid category value")

    async def _bcp_receive_monitor_stop(self, client, category):
        """Stop monitoring the specified category."""
        category = str.lower(category)

        if category == "events":
            self._monitor_events_stop(client)
        elif category == "devices":
            self._monitor_devices_stop(client)
        elif category == "drivers":
            self._monitor_drivers_stop(client)
        elif category == "switches":
            self._monitor_switches_stop(client)
        elif category == "machine_vars":
            self._monitor_machine_vars_stop(client)
        elif category == "player_vars":
            self._monitor_player_vars_stop(client)
        elif category == "modes":
            self._monitor_modes_stop(client)
        elif category == "core_events":
            self._monitor_core_events_stop(client)
        elif category == "status_request":
            self._monitor_status_request_stop(client)
        else:
            self.machine.bcp.transport.send_to_client(client,
                                                      "error",
                                                      cmd="monitor_stop?category={}".format(category),
                                                      error="Invalid category value")

    def _monitor_drivers(self, client):
        """Monitor all drivers."""
        self.machine.bcp.transport.add_handler_to_transport("_monitor_drivers", client)

    def _monitor_drivers_stop(self, client):
        """Monitor all drivers."""
        self.machine.bcp.transport.remove_transport_from_handle("_monitor_drivers", client)

    def _monitor_events(self, client):
        """Monitor all events."""
        self.machine.bcp.transport.add_handler_to_transport("_monitor_events", client)
        self.machine.events.monitor_events = True

    def _monitor_events_stop(self, client):
        """Stop monitoring all events for the specified client."""
        self.machine.bcp.transport.remove_transport_from_handle("_monitor_events", client)

        if not self.machine.bcp.transport.get_transports_for_handler("_monitor_events"):
            self.machine.events.monitor_events = False

    def monitor_posted_event(self, posted_event: PostedEvent):
        """Send monitored posted event to bcp clients."""
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_monitor_events",
            bcp_command="monitored_event",
            event_name=posted_event.event,
            event_type=posted_event.type,
            event_callback=posted_event.callback,
            event_kwargs=Util.convert_to_simply_type(posted_event.kwargs),
            registered_handlers=Util.convert_to_simply_type(
                self.machine.events.registered_handlers.get(posted_event.event, []))
        )

    def _monitor_devices(self, client):
        """Register client to get notified of device changes."""
        self.machine.bcp.transport.add_handler_to_transport("_devices", client)
        # trigger updates of lights
        self.machine.light_controller.monitor_lights()

        # initially send all states
        for collection in self.machine.device_manager.get_monitorable_devices().values():
            for device in collection.values():
                self.machine.bcp.transport.send_to_client(
                    client=client,
                    bcp_command='device',
                    type=device.class_label,
                    name=device.name,
                    changes=False,
                    state=device.get_monitorable_state())

    def _monitor_devices_stop(self, client):
        """Remove client to no longer get notified of device changes."""
        self.machine.bcp.transport.remove_transport_from_handle("_devices", client)

    def notify_device_changes(self, device, attribute_name, old_value, new_value):
        """Notify all listeners about device change."""
        if not self.configured:
            return

        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_devices",
            bcp_command='device',
            type=device.class_label,
            name=device.name,
            changes=(attribute_name, Util.convert_to_simply_type(old_value), Util.convert_to_simply_type(new_value)),
            state=device.get_monitorable_state())

    def _monitor_switches(self, client):
        """Register client to get notified of switch changes."""
        self.machine.switch_controller.add_monitor(self._notify_switch_changes)
        self.machine.bcp.transport.add_handler_to_transport("_switches", client)

    def _monitor_switches_stop(self, client):
        """Remove client to no longer get notified of switch changes."""
        self.machine.bcp.transport.add_handler_to_transport("_switches", client)

        # If there are no more clients monitoring switches, remove monitor
        if not self.machine.bcp.transport.get_transports_for_handler("_switches"):
            self.machine.switch_controller.remove_monitor(self._notify_switch_changes)

    def _notify_switch_changes(self, change: MonitoredSwitchChange):
        """Notify all listeners about switch change."""
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_switches",
            bcp_command='switch',
            name=change.name,
            state=change.state)

    def _monitor_player_vars(self, client):
        # Setup player variables to be monitored (if necessary)
        if not self.machine.bcp.transport.get_transports_for_handler("_player_vars"):
            Player.monitor_enabled = True
            self.machine.register_monitor('player', self._player_var_change)

        self.machine.bcp.transport.add_handler_to_transport("_player_vars", client)

    def _monitor_player_vars_stop(self, client):
        self.machine.bcp.transport.remove_transport_from_handle("_player_vars", client)

        # If there are no more clients monitoring player variables, stop monitoring
        if not self.machine.bcp.transport.get_transports_for_handler("_player_vars"):
            Player.monitor_enabled = False

    def _monitor_machine_vars(self, client):
        # Setup machine variables to be monitored (if necessary)
        if not self.machine.bcp.transport.get_transports_for_handler("_machine_vars"):
            self.machine.variables.machine_var_monitor = True
            self.machine.register_monitor('machine_vars', self._machine_var_change)

        # Send initial machine variable values
        self._send_machine_vars(client)

        # Establish handler for machine variable changes
        self.machine.bcp.transport.add_handler_to_transport("_machine_vars", client)

    def _monitor_machine_vars_stop(self, client):
        self.machine.bcp.transport.remove_transport_from_handle("_machine_vars", client)

        # If there are no more clients monitoring machine variables, stop monitoring
        if not self.machine.bcp.transport.get_transports_for_handler("_machine_vars"):
            self.machine.machine_var_monitor = False

    def _send_machine_vars(self, client):
        self.machine.bcp.transport.send_to_client(
            client, bcp_command='settings', settings=Util.convert_to_simply_type(self.machine.settings.get_settings()))
        for var_name, settings in self.machine.variables.machine_vars.items():
            self.machine.bcp.transport.send_to_client(client, bcp_command='machine_variable',
                                                      name=var_name,
                                                      value=settings['value'])

    # pylint: disable-msg=too-many-arguments
    def _player_var_change(self, name, value, prev_value, change, player_num):
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_player_vars",
            bcp_command='player_variable',
            name=name,
            value=value,
            prev_value=prev_value,
            change=change,
            player_num=player_num)

    def _machine_var_change(self, name, value, prev_value, change):
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_machine_vars",
            bcp_command='machine_variable',
            name=name,
            value=value,
            prev_value=prev_value,
            change=change)

    def _monitor_modes(self, client):
        """Begin monitoring all mode events (start, stop) via the specified client."""
        if not self.machine.bcp.transport.get_transports_for_handler("_modes"):
            self.machine.mode_controller.register_start_method(self._mode_start, 'mode')
            self.machine.events.add_handler("modes_active_modes_changed", self._send_mode_list)

        self.machine.bcp.transport.add_handler_to_transport("_modes", client)

        self.machine.bcp.transport.send_to_client(
            client=client,
            bcp_command="mode_list",
            running_modes=[(m.name, m.priority) for m in self.machine.mode_controller.active_modes])

    def _send_mode_list(self, **kwargs):
        """Send list of current modes."""
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_modes",
            bcp_command="mode_list",
            running_modes=[(m.name, m.priority) for m in self.machine.mode_controller.active_modes])

    def _monitor_modes_stop(self, client):
        """Stop monitoring all mode events (start, stop) via the specified client."""
        self.machine.bcp.transport.remove_transport_from_handle("_modes", client)

        if not self.machine.bcp.transport.get_transports_for_handler("_modes"):
            self.machine.mode_controller.remove_start_method(self._mode_start, 'mode')
            self.machine.events.remove_handler_by_event("modes_active_modes_changed", self._send_mode_list)

    def _mode_start(self, config, priority, mode, **kwargs):
        """Send 'mode_start' to the monitoring clients."""
        del config
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_modes",
            bcp_command="mode_start",
            name=mode.name,
            priority=priority)

        # Return the method and mode name to call when the mode stops (self-registering)
        return self._mode_stop, mode.name

    def _mode_stop(self, mode, **kwargs):
        """Send 'mode_stop' to the monitoring clients."""
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_modes",
            bcp_command="mode_stop",
            name=mode)

    def _monitor_core_events(self, client):
        """Begin monitoring all core events (ball, player turn, etc.) via the specified client."""
        if not self.machine.bcp.transport.get_transports_for_handler("_core_events"):
            self.machine.events.add_handler('ball_started', self._ball_started)
            self.machine.events.add_handler('ball_ended', self._ball_ended)
            self.machine.events.add_handler('player_turn_started', self._player_turn_start)
            self.machine.events.add_handler('player_added', self._player_added)

        self.machine.bcp.transport.add_handler_to_transport("_core_events", client)

    def _monitor_core_events_stop(self, client):
        """Stop monitoring all core events (ball, player turn, etc.) via the specified client."""
        self.machine.bcp.transport.remove_transport_from_handle("_core_events", client)

        if not self.machine.bcp.transport.get_transports_for_handler("_core_events"):
            self.machine.events.remove_handler_by_event('ball_started', self._ball_started)
            self.machine.events.remove_handler_by_event('ball_ended', self._ball_ended)
            self.machine.events.remove_handler_by_event('player_turn_started', self._player_turn_start)
            self.machine.events.remove_handler_by_event('player_added', self._player_added)

    def _monitor_status_request(self, client):
        """Begin monitoring status_request messages via the specified client."""
        self.machine.bcp.transport.add_handler_to_transport("_status_request", client)

    def _monitor_status_request_stop(self, client):
        """Stop monitoring status_request messages via the specified client."""
        self.machine.bcp.transport.remove_transport_from_handle("_status_request", client)

    def _ball_started(self, ball, player, **kwargs):
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_core_events",
            bcp_command="ball_start",
            player_num=player,
            ball=ball)

    def _ball_ended(self, **kwargs):
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_core_events",
            bcp_command="ball_end")

    def _player_turn_start(self, number, player, **kwargs):
        del player
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_core_events",
            bcp_command="player_turn_start",
            player_num=number)

    def _player_added(self, num, player, **kwargs):
        del player
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler(
            handler="_core_events",
            bcp_command="player_added",
            player_num=num)

    async def process_bcp_message(self, cmd, kwargs, client):
        """Process BCP message.

        Args:
        ----
            cmd: The command for this message.
            kwargs: Arguments for the command.
            client: Client which send this message.
        """
        if self._debug_to_console or self._debug_to_file:
            if 'rawbytes' in kwargs:
                debug_kwargs = deepcopy(kwargs)
                debug_kwargs['rawbytes'] = '<{} bytes>'.format(
                    len(debug_kwargs.pop('rawbytes')))

                self.debug_log("Processing command: %s %s", cmd, debug_kwargs)
            else:
                self.debug_log("Processing command: %s %s", cmd, kwargs)

        if cmd in self.bcp_receive_commands:
            try:
                callback = self.bcp_receive_commands[cmd]
            except TypeError as e:
                self.machine.bcp.transport.send_to_client(client, "error", cmd=cmd, error=str(e), kwargs=kwargs)
            else:
                await callback(client=client, **kwargs)

        else:
            self.warning_log("Received invalid BCP command: %s from client: %s", cmd, client.name)

    async def _bcp_receive_error(self, client, **kwargs):
        """Handle a BCP error message from a remote BCP host indicating that a command from MPF was not recognized.

        This method only posts a warning to the log. It doesn't do anything else
        at this point.
        """
        self.warning_log('Received Error command from host with parameters: %s, from client %s',
                         kwargs, str(client))

    def send_driver_event(self, **kwargs):
        """Notify all observers about driver event."""
        self.machine.bcp.transport.send_to_clients_with_handler("_monitor_drivers", "driver_event", **kwargs)

    async def _bcp_receive_reset_complete(self, client, **kwargs):
        """Handle a BCP reset_complete message from a remote BCP host indicating their reset process has completed."""
        del kwargs
        self.debug_log("Received reset_complete from client: %s %s", client.name)
        self._client_reset_complete_status[client] = True

        # Check if reset_complete status is True from all clients
        if all(status is True for item, status in self._client_reset_complete_status.items()):
            if self._client_reset_queue:
                self._client_reset_queue.clear()
                self._client_reset_queue = None
            self._client_reset_complete_status.clear()
            self.debug_log("Received reset_complete from all clients. Clearing wait from queue event.")

    def bcp_reset(self, queue, **kwargs):
        """Send the 'reset' command to the remote BCP host."""
        del kwargs

        # Will hold the queue event until all clients respond with a "reset_complete" command
        clients = self.machine.bcp.transport.get_all_clients()
        self._client_reset_complete_status.clear()
        for client in clients:
            if not client.name:
                continue
            self._client_reset_complete_status[client] = False

        if self._client_reset_complete_status:
            queue.wait()
            self._client_reset_queue = queue

            # Send the reset command
            self.debug_log("Sending reset to all clients (will now wait for reset_complete "
                           "to be received from all clients).")
            self.machine.bcp.transport.send_to_all_clients("reset")

    async def _bcp_receive_switch(self, client, name, state, **kwargs):
        """Process an incoming switch state change request from a remote BCP host.

        Args:
        ----
            client: Client which sent the switch state.
            name: String name of the switch to set.
            state: Integer representing the state this switch will be set to.
                1 = active, 0 = inactive, -1 means this switch will be flipped
                from whatever its current state is to the opposite state.
            kwargs: Additional arguments (unused)
        """
        del kwargs
        del client
        state = int(state)

        try:
            switch = self.machine.switches[name]
        except KeyError:
            self.warning_log("Received BCP switch message with invalid switch"
                             "name: '%s'", name)
            return

        if state == -1:
            if self.machine.switch_controller.is_active(switch):
                state = 0
            else:
                state = 1

        self.machine.switch_controller.process_switch_obj(obj=switch, state=state, logical=True)

    async def _evaluate_placeholder(self, client, placeholder, parameters=None, **kwargs):
        """Evaluate and return placeholder."""
        del kwargs
        if parameters is None:
            parameters = []
        placeholder_obj = self.machine.placeholder_manager.build_raw_template(placeholder, None)
        try:
            value = placeholder_obj.evaluate(parameters=parameters)
        except AssertionError as e:
            self.machine.bcp.transport.send_to_client(client=client, bcp_command='evaluate_placeholder',
                                                      error=str(e))
            return

        self.machine.bcp.transport.send_to_client(client=client, bcp_command='evaluate_placeholder', value=value,
                                                  error=False)

    async def _bcp_receive_register_trigger(self, client, event, **kwargs):
        """Register a trigger for a client."""
        del kwargs
        self.add_registered_trigger_event_for_client(client, event)

    async def _bcp_receive_deregister_trigger(self, client, event, **kwargs):
        """Deregister a trigger for a client."""
        del kwargs
        self.remove_registered_trigger_event_for_client(client, event)

    def bcp_player_added(self, num, **kwargs):
        """Send BCP 'player_added' to the connected BCP hosts."""
        del kwargs
        self.machine.bcp.transport.send_to_clients_with_handler('_player_vars', 'player_added', player_num=num)

    def bcp_trigger(self, name, **kwargs):
        """Send BCP 'trigger' to the connected BCP hosts."""
        # ignore events which already came from bcp to prevent loops
        if "_from_bcp" in kwargs:
            return

        # Since player variables are sent automatically, if we get a trigger
        # for an event that starts with "player_", we need to only send it here
        # if there's *not* a player variable with that name, since if there is
        # a player variable then the player variable handler will send it.
        if name.startswith('player_'):
            try:
                if self.machine.game.player.is_player_var(name.lstrip('player_')):
                    return

            except AttributeError:
                pass

        self.machine.bcp.transport.send_to_clients_with_handler(
            handler=name, bcp_command='trigger', name=name, **kwargs)

    def bcp_trigger_client(self, client, name, **kwargs):
        """Send BCP 'trigger' to a specific client."""
        # ignore events which already came from bcp to prevent loops
        if "_from_bcp" in kwargs:
            return

        self.machine.bcp.transport.send_to_client(client=client, bcp_command='trigger', name=name, **kwargs)

    async def _bcp_receive_trigger(self, client, name, callback=None, **kwargs):
        """Process an incoming trigger command from a remote BCP host."""
        del client
        kwargs['_from_bcp'] = True
        if callback:
            self.machine.events.post(event=name,
                                     callback=self.bcp_trigger,
                                     name=kwargs.pop('callback'),
                                     **kwargs)

        else:
            self.machine.events.post(event=name, **kwargs)



from mpf.tests.MpfTestCase import MpfTestCase


class TestBallDeviceModernTroughPlungerSetup(MpfTestCase):

    def get_platform(self):
        return "smart_virtual"

    def get_config_file(self):
        return 'test_modern_trough_plunger_setup.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_ball_in_plunger_during_eject(self):
        # add some initial balls
        self.hit_switch_and_run("s_trough_switch1", 0)
        self.hit_switch_and_run("s_trough_switch2", 0)
        self.hit_switch_and_run("s_trough_switch3", 1)
        self.assertNumBallsKnown(3)

        self.machine.playfield.add_ball()
        self.hit_switch_and_run("s_ball_switch_plunger_lane", 0)
        self.advance_time_and_run(100)

        self.assertEqual("idle", self.machine.ball_devices["bd_trough"].state)
        self.assertEqual("idle", self.machine.ball_devices["bd_plunger"].state)
        self.assertBallsOnPlayfield(2)
        self.assertNumBallsKnown(4)


class TestBallDeviceModernTroughPlungerSetupRaces(MpfTestCase):

    def get_platform(self):
        return "virtual"

    def get_config_file(self):
        return 'test_modern_trough_plunger_setup.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_ball_in_plunger_during_eject(self):
        # add some initial balls
        self.mock_event("balldevice_bd_trough_ejecting_ball")
        self.mock_event("balldevice_bd_plunger_ejecting_ball")
        self.mock_event("balldevice_bd_trough_ball_eject_success")
        self.mock_event("balldevice_bd_plunger_ball_eject_success")
        self.hit_switch_and_run("s_trough_switch1", 0)
        self.hit_switch_and_run("s_trough_switch2", 0)
        self.hit_switch_and_run("s_trough_switch3", 1)
        self.advance_time_and_run(10)
        self.assertNumBallsKnown(3)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(.3)
        self.assertEventCalled("balldevice_bd_trough_ejecting_ball")
        self.mock_event("balldevice_bd_trough_ejecting_ball")
        self.assertEventNotCalled("balldevice_bd_plunger_ejecting_ball")
        self.release_switch_and_run("s_trough_switch3", 1)

        self.hit_switch_and_run("s_ball_switch_plunger_lane", 0)
        self.advance_time_and_run(10)
        self.assertEventNotCalled("balldevice_bd_trough_ejecting_ball")
        self.assertEventCalled("balldevice_bd_plunger_ejecting_ball")
        self.mock_event("balldevice_bd_plunger_ejecting_ball")

        self.release_switch_and_run("s_ball_switch_plunger_lane", 0)
        self.advance_time_and_run(3)

        self.assertEventNotCalled("balldevice_bd_trough_ejecting_ball")
        self.assertEventNotCalled("balldevice_bd_plunger_ejecting_ball")

        self.hit_and_release_switch("s_playfield")
        self.advance_time_and_run(20)

        self.assertEventNotCalled("balldevice_bd_trough_ejecting_ball")
        self.assertEventNotCalled("balldevice_bd_plunger_ejecting_ball")
        self.assertEventCalled("balldevice_bd_trough_ball_eject_success", 1)
        self.assertEventCalled("balldevice_bd_plunger_ball_eject_success", 1)

        self.assertEqual("idle", self.machine.ball_devices["bd_trough"].state)
        self.assertEqual("idle", self.machine.ball_devices["bd_plunger"].state)
        self.assertBallsOnPlayfield(1)
        self.assertNumBallsKnown(3)

    def test_eject_failure_with_jam(self):
        """Trough jams but pluger confirms eject."""
        # add some initial balls
        self.mock_event("balldevice_bd_trough_ejecting_ball")
        self.mock_event("balldevice_bd_plunger_ejecting_ball")
        self.mock_event("balldevice_bd_trough_ball_eject_success")
        self.mock_event("balldevice_bd_trough_ball_eject_failed")
        self.mock_event("balldevice_bd_plunger_ball_eject_success")
        self.hit_switch_and_run("s_trough_switch1", 0)
        self.hit_switch_and_run("s_trough_switch2", 0)
        self.hit_switch_and_run("s_trough_switch3", 1)
        self.advance_time_and_run(10)
        self.assertNumBallsKnown(3)

        self.machine.playfield.add_ball()
        self.advance_time_and_run(.1)
        self.assertEventCalled("balldevice_bd_trough_ejecting_ball")
        self.mock_event("balldevice_bd_trough_ejecting_ball")
        self.assertEventNotCalled("balldevice_bd_plunger_ejecting_ball")
        self.hit_switch_and_run("s_trough_jam", 0)
        self.release_switch_and_run("s_trough_switch3", .25)
        self.hit_switch_and_run("s_trough_switch3", .25)
        self.release_switch_and_run("s_trough_switch1", 0)

        # eject timeout not passed yet
        self.advance_time_and_run(1.51)
        self.assertEventNotCalled("balldevice_bd_trough_ball_eject_failed")
        self.assertEventNotCalled("balldevice_bd_trough_ball_eject_success")

        # unexpected playfield activation happens (i.e. broken switch)
        self.hit_and_release_switch("s_playfield")

        # event timeout passes
        self.advance_time_and_run(1)
        self.assertEventCalled("balldevice_bd_trough_ball_eject_failed")
        self.assertEventNotCalled("balldevice_bd_trough_ball_eject_success")

        self.hit_and_release_switch("s_playfield")

        self.assertEqual("ejecting", self.machine.ball_devices["bd_trough"].state)
        self.assertEqual("waiting_for_ball", self.machine.ball_devices["bd_plunger"].state)

        self.assertEventCalled("balldevice_bd_trough_ejecting_ball", 1)
        self.release_switch_and_run("s_trough_jam", 0)

        self.advance_time_and_run(20)

        self.assertEqual("idle", self.machine.ball_devices["bd_trough"].state)
        self.assertEqual("idle", self.machine.ball_devices["bd_plunger"].state)

        self.assertBallsOnPlayfield(1)
        self.assertNumBallsKnown(3)



from mpf.tests.MpfTestCase import MpfTestCase


class TestTimedSwitch(MpfTestCase):

    def get_config_file(self):
        return 'timed_switches.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/timed_switches/'

    def test_in_mode(self):
        self.start_mode("mode1")
        self.mock_event('mode_switch_active')
        self.mock_event('mode_switch_released')

        self.hit_switch_and_run("switch2", 1.1)
        self.assertEventNotCalled("mode_switch_active")
        self.advance_time_and_run()
        self.assertEventCalled("mode_switch_active")

    def test_timed_switches(self):

        # test single switch

        self.mock_event('group1_active')
        self.mock_event('group1_released')

        self.hit_switch_and_run('switch1', 1)
        self.assertEventNotCalled('group1_active')
        self.assertEventNotCalled('group1_released')

        self.advance_time_and_run(1.1)
        self.assertEventCalled('group1_active')
        self.assertEventNotCalled('group1_released')

        self.advance_time_and_run(10)
        self.assertEventCalled('group1_active', 1)
        self.assertEventNotCalled('group1_released')

        self.release_switch_and_run('switch1', .1)
        self.assertEventCalled('group1_released')

        # test 2 switches together
        self.mock_event('group1_active')
        self.mock_event('group1_released')

        self.hit_switch_and_run('switch1', 1)
        self.assertEventNotCalled('group1_active')
        self.assertEventNotCalled('group1_released')

        self.advance_time_and_run(1.1)
        self.assertEventCalled('group1_active')
        self.assertEventNotCalled('group1_released')

        self.hit_switch_and_run('switch2', 3)
        self.assertEventCalled('group1_active')
        self.assertEventNotCalled('group1_released')

        self.release_switch_and_run('switch1', .1)
        self.assertEventCalled('group1_active', 1)
        self.assertEventNotCalled('group1_released')

        self.release_switch_and_run('switch2', .1)
        self.assertEventCalled('group1_active', 1)
        self.assertEventCalled('group1_released')

        # test inverted state & custom events

        self.mock_event('active_event')
        self.mock_event('release_event')

        self.hit_switch_and_run('switch3', 5)
        self.assertEventNotCalled('active_event')
        self.assertEventNotCalled('release_event')

        self.release_switch_and_run('switch3', 1)
        self.assertEventNotCalled('active_event')
        self.assertEventNotCalled('release_event')

        self.advance_time_and_run(2)
        self.assertEventCalled('active_event')
        self.assertEventNotCalled('release_event')

        self.hit_switch_and_run('switch3', .1)
        self.assertEventCalled('active_event')
        self.assertEventCalled('release_event')

        # test built-in flipper cradle (also tests tags)

        self.mock_event('flipper_cradle')
        self.mock_event('flipper_cradle_release')

        self.hit_switch_and_run('switch4', 1)
        self.assertEventNotCalled('flipper_cradle')
        self.assertEventNotCalled('flipper_cradle_release')

        self.advance_time_and_run(3)
        self.assertEventCalled('flipper_cradle')
        self.assertEventNotCalled('flipper_cradle_release')

        # tap the second flipper
        self.hit_and_release_switch('switch5')
        self.assertEventNotCalled('flipper_cradle_release')

        # release the initial switch
        self.release_switch_and_run('switch4', 1)
        self.assertEventCalled('flipper_cradle_release')



"""Command to start the MPF game engine."""

import argparse

from ruamel.yaml import YAML

from mpf.core.machine import MachineController
from mpf.core.utility_functions import Util
from mpf.core.config_loader import YamlMultifileConfigLoader
from mpf.wire.fast import wire


class Command:

    """Runs the mpf game."""

    # pylint: disable-msg=too-many-locals,too-many-statements
    def __init__(self, mpf_path, machine_path, args):
        """Generate wiring for game from MPL file."""
        del mpf_path
        self.machine = None
        self._sigint_count = 0

        parser = argparse.ArgumentParser(description='Generates wiring .yaml file')

        parser.add_argument("-c",
                            action="store", dest="configfile",
                            default="config.yaml", metavar='config_file',
                            help="The name of a config file to load. Default "
                                 "is "
                                 "config.yaml. Multiple files can be used "
                                 "via a comma-"
                                 "separated list (no spaces between)")

        self.args = parser.parse_args(args)
        self.args.configfile = Util.string_to_event_list(self.args.configfile)

        # To initialize and check machine, load it onto the virtual platform - we have to use the virtual platform
        # because it would be helpful to be able to calculate wiring before setting up physical hardware.
        self.args.__dict__["production"] = False
        self.args.__dict__["force_platform"] = "smart_virtual"
        self.args.__dict__["text_ui"] = False
        self.args.__dict__["bcp"] = False

        config_loader = YamlMultifileConfigLoader(machine_path, self.args.configfile,
                                                  False, False)

        config = config_loader.load_mpf_config()

        # print(config.get_machine_config())

        self.machine = MachineController(vars(self.args), config)
        self.machine.initialise_mpf()

        result = wire(self.machine)

        yaml = YAML()
        yaml.default_flow_style = False
        f = open("wiring.yaml", "w", encoding="utf-8")
        yaml.dump(result, f)
        f.close()



"""WS2812 LED on the FAST controller."""
import logging

from typing import Optional
from typing import List

from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface


class FASTDirectLED:

    """FAST RGB LED."""

    __slots__ = ["number", "number_int", "dirty", "hardware_fade_ms", "log", "channels", "machine"]

    def __init__(self, number: str, hardware_fade_ms: int, machine) -> None:
        """Initialise FAST LED."""
        self.number_int = int(number)
        self.number = Util.int_to_hex_string(self.number_int)
        self.dirty = True
        self.machine = machine
        self.hardware_fade_ms = hardware_fade_ms
        self.log = logging.getLogger('FASTLED')
        self.channels = [None, None, None]      # type: List[Optional[FASTDirectLEDChannel]]
        # All FAST LEDs are 3 element RGB and are set using hex strings
        self.log.debug("Creating FAST RGB LED at hardware address: %s", self.number)

    def add_channel(self, channel_num: int, channel_obj: "FASTDirectLEDChannel"):
        """Add channel to LED."""
        self.channels[channel_num] = channel_obj

    @property
    def current_color(self):
        """Return current color."""
        result = ""
        self.dirty = False
        current_time = self.machine.clock.get_time()
        # send this as grb because the hardware will twist it again
        for index in [1, 0, 2]:
            channel = self.channels[index]
            if channel:
                brightness, _, done = channel.get_fade_and_brightness(current_time)
                result += hex(int(brightness * 255))[2:].zfill(2)
                if not done:
                    self.dirty = True
            else:
                result += "00"

        return result


class FASTDirectLEDChannel(LightPlatformInterface):

    """Represents a single RGB LED channel connected to the Fast hardware platform."""

    __slots__ = ["led", "channel", "_current_fade", "_last_brightness"]

    def __init__(self, led: FASTDirectLED, channel) -> None:
        """Initialise LED."""
        super().__init__("{}-{}".format(led.number, channel))
        self.led = led
        self.channel = int(channel)
        self._current_fade = (0, -1, 0, -1)
        self._last_brightness = None

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Set brightness via callback."""
        self.led.dirty = True
        self._current_fade = (start_brightness, start_time, target_brightness, target_time)
        self._last_brightness = None

    def get_fade_and_brightness(self, current_time):
        """Return fade + brightness and mark as clean if this is it."""
        if self._last_brightness:
            return self._last_brightness, 0, True
        max_fade_ms = self.led.hardware_fade_ms
        start_brightness, start_time, target_brightness, target_time = self._current_fade
        fade_ms = int((target_time - current_time) * 1000.0)
        if fade_ms > max_fade_ms > 0:
            fade_ms = max_fade_ms
            ratio = ((current_time + (fade_ms / 1000.0) - start_time) /
                     (target_time - start_time))
            brightness = start_brightness + (target_brightness - start_brightness) * ratio
            done = False
        else:
            if fade_ms < 0:
                fade_ms = 0
            brightness = target_brightness
            self._last_brightness = brightness
            done = True

        return brightness, fade_ms, done

    def get_board_name(self):
        """Return the board of this light."""
        return "FAST LED CPU"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.led.number_int * 3 + self.channel == other.led.number_int * 3 + other.channel + 1

    def get_successor_number(self):
        """Return next number."""
        if self.channel == 2:
            return "{}-0".format(self.led.number_int + 1)

        return "{}-{}".format(self.led.number_int, self.channel + 1)

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return (self.led.number_int, self.channel) < (other.led.number_int, other.channel)



"""Contains the base class for digital (image-based) score reels."""

from mpf.core.system_wide_device import SystemWideDevice


class DigitalScoreReel(SystemWideDevice):

    """A digital score reel."""

    config_section = 'digital_score_reels'
    collection = 'digital_score_reels'
    class_label = 'digital_score_reel'

    def __init__(self, machine, name):
        """Initialize digital score reel."""
        super().__init__(machine, name)
        self._frames = {}
        self._reel_count = 0
        self._include_player_number = False

    async def _initialize(self):
        await super()._initialize()

        self._reel_count = self.config["reel_count"]
        self._include_player_number = self.config["include_player_number"]
        for frame in self.config["frames"]:
            self._frames[str(frame["character"])] = str(frame["frame"])

        self.machine.events.add_handler(self.name, self._post_reel_values)

    def _post_reel_values(self, **kwargs):
        # Pad the string up to the necessary number of characters in the reel
        score = str(kwargs["value"]).rjust(self._reel_count, self.config["start_value"])
        # Create a dict of reel name keys to target frame values
        result = {str(i + 1): self._frames[score[i]] for i in range(self._reel_count)}
        # Post the event
        event_name = "score_reel_{}_player{}".format(self.name, self.machine.game.player.number) if \
            self._include_player_number else "score_reel_{}".format(self.name)
        self.machine.events.post(event_name, **result)



"""Support for i2c devices on a bus."""
import abc
from typing import Any


class I2cPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for a i2c device on a bus in hardware platforms."""

    __slots__ = ["number"]

    def __init__(self, number: Any) -> None:
        """Remember the number."""
        self.number = number

    @abc.abstractmethod
    def i2c_write8(self, register, value):
        """Write an 8-bit value to a specific register via I2C.

        Args:
        ----
            register (int): Register
            value (int): Value to write
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def i2c_read_block(self, register, count):
        """Read an len bytes from an register via I2C.

        Args:
        ----
            register (int): Register
            count (int): Bytes to read
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def i2c_read8(self, register):
        """Read an 8-bit value from an register via I2C.

        Args:
        ----
            register (int): Register
        """
        raise NotImplementedError



"""P-Roc hardware platform devices."""
from typing import Tuple, Optional

import asyncio
import logging

from mpf.core.platform_batch_light_system import PlatformBatchLight
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface

from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface

from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.core.utility_functions import Util

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.platforms.p_roc_common import PROCBasePlatform     # pylint: disable-msg=cyclic-import,unused-import


class PROCSwitch(SwitchPlatformInterface):

    """P-ROC switch object which is use to store the configure rules and config."""

    __slots__ = ["string_number", "log", "notify_on_nondebounce", "hw_rules", "pdbconfig"]

    def __init__(self, config, number, notify_on_nondebounce, platform):
        """Initialise P-ROC switch."""
        super().__init__(config, number)
        self.string_number = number
        self.log = logging.getLogger('PROCSwitch {}'.format(self.string_number))
        self.notify_on_nondebounce = notify_on_nondebounce
        self.hw_rules = {"closed_debounced": [],
                         "closed_nondebounced": [],
                         "open_debounced": [],
                         "open_nondebounced": []}
        self.pdbconfig = getattr(platform, "pdbconfig", None)

    def get_board_name(self):
        """Return board of the switch."""
        if not self.pdbconfig:
            return "P-Roc"

        board, bank, _ = self.pdbconfig.decode_pdb_address(self.string_number)

        return "SW-16 Board {} Bank {}".format(board, bank)

    @property
    def has_rules(self):
        """Return true as we support hardware rules."""
        return True


class PROCDriver(DriverPlatformInterface):

    """A P-ROC driver/coil.

    Base class for drivers connected to a P3-ROC. This class is used for all
    drivers, regardless of whether they're connected to a P-ROC driver board
    (such as the PD-16 or PD-8x8) or an OEM driver board.

    """

    __slots__ = ["log", "proc", "string_number", "pdbconfig", "polarity", "__dict__"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, number, config, platform, string_number, polarity):
        """Initialise driver."""
        self.log = logging.getLogger('PROCDriver {}'.format(number))
        super().__init__(config, number)
        self.platform = platform            # type: PROCBasePlatform
        self.polarity = polarity
        self.string_number = string_number
        self.pdbconfig = getattr(platform, "pdbconfig", None)

    async def initialise(self):
        """Initialize driver."""
        if self.polarity is None:
            self.log.debug("Getting polarity for driver %s", self.number)
            state = await self.platform.run_proc_cmd("driver_get_state", self.number)
            self.polarity = state["polarity"]
            self.log.debug("Got Polarity %s for driver %s", self.polarity, self.number)

        self.log.debug("Driver Settings for %s", self.number)
        self.platform.run_proc_cmd_no_wait("driver_update_state", self.state())

    def get_board_name(self):
        """Return board of the driver."""
        if not self.pdbconfig:
            return "P-Roc"

        board, bank, _ = self.pdbconfig.decode_pdb_address(self.string_number)

        return "PD-16 Board {} Bank {}".format(board, bank)

    @classmethod
    def get_pwm_on_off_ms(cls, coil: HoldSettings):
        """Find out the pwm_on_ms and pwm_off_ms for this driver."""
        return Util.power_to_on_off(coil.power)

    def disable(self):
        """Disable (turn off) this driver."""
        self.log.debug('Disabling Driver')
        self.platform.run_proc_cmd_no_wait("driver_disable", self.number)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable (turn on) this driver."""
        if pulse_settings.power != 1:
            raise AssertionError("pulse_power != 1.0 with hold/enable is not supported in P/P3-Roc.")

        if hold_settings.power < 1.0:
            pwm_on, pwm_off = self.get_pwm_on_off_ms(hold_settings)
            self.log.debug('Enabling. Initial pulse_ms:%s, pwm_on_ms: %s'
                           'pwm_off_ms: %s', pwm_on, pwm_off, pulse_settings.duration)

            self.platform.run_proc_cmd_no_wait("driver_patter", self.number, pwm_on, pwm_off,
                                               pulse_settings.duration, True)
        else:
            self.log.debug('Enabling at 100%')

            self.platform.run_proc_cmd_no_wait("driver_schedule", self.number, 0xffffffff, 0, True)

    def pulse(self, pulse_settings: PulseSettings):
        """Enable this driver for `milliseconds`.

        ``ValueError`` will be raised if `milliseconds` is outside of the range
        0-255.
        """
        # make sure we never set 0 (due to a bug elsewhere) as this would turn the driver on permanently
        assert pulse_settings.duration != 0

        self.log.debug('Pulsing for %sms at %s pulse power.', pulse_settings.duration, pulse_settings.power)
        if pulse_settings.power != 1:
            on_time, off_time = Util.power_to_on_off(pulse_settings.power)
            self.platform.run_proc_cmd_no_wait("driver_pulsed_patter", self.number, on_time, off_time,
                                               pulse_settings.duration, True)
        else:
            self.platform.run_proc_cmd_no_wait("driver_pulse", self.number, pulse_settings.duration)

    def state(self):
        """Return a dictionary representing this driver's current configuration state."""
        return {
            'driverNum': self.number,
            'outputDriveTime': 0,
            'polarity': self.polarity,
            'state': False,
            'waitForFirstTimeSlot': False,
            'timeslots': 0,
            'patterOnTime': 0,
            'patterOffTime': 0,
            'patterEnable': False,
            'futureEnable': False
        }

    @property
    def has_rules(self):
        """Return true as we support hardware rules."""
        return True


class PROCMatrixLight(LightPlatformSoftwareFade):

    """A P-ROC matrix light device."""

    __slots__ = ["log", "proc", "platform"]

    def __init__(self, number, machine, platform):
        """Initialise matrix light device."""
        super().__init__(number, machine.clock.loop,
                         int(1 / machine.config['mpf']['default_light_hw_update_hz'] * 1000))
        self.log = logging.getLogger('PROCMatrixLight {}'.format(number))
        self.platform = platform

    def set_brightness(self, brightness: float):
        """Enable (turns on) this driver."""
        if brightness >= 1:
            self.platform.run_proc_cmd_no_wait("driver_schedule", self.number, 0xffffffff, 0, True)
        elif brightness > 0:
            pwm_on_ms, pwm_off_ms = Util.power_to_on_off(brightness)
            self.platform.run_proc_cmd_no_wait("driver_patter", self.number, pwm_on_ms, pwm_off_ms, 0, True)
        else:
            self.platform.run_proc_cmd_no_wait("driver_disable", self.number)

    def get_board_name(self):
        """Return board of the light."""
        return "P-Roc Matrix"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        return self.number == other.number + 1

    def get_successor_number(self):
        """Return next number."""
        return self.number + 1

    def __lt__(self, other):
        """Order lights by hardware number."""
        return self.number < other.number


class PDBSwitch:

    """Base class for switches connected to a P-ROC/P3-ROC."""

    def __init__(self, pdb, number_str):
        """Find out the number of the switch."""
        upper_str = number_str.upper()
        if upper_str.startswith('SD'):  # only P-ROC
            self.sw_number = int(upper_str[2:])
        elif upper_str.count("/") == 1:  # only P-ROC
            self.sw_number = self.parse_matrix_num(upper_str)
        else:   # only P3-Roc
            try:
                (boardnum, banknum, inputnum) = pdb.decode_pdb_address(number_str)
                self.sw_number = boardnum * 16 + banknum * 8 + inputnum
            except ValueError:
                try:
                    self.sw_number = int(number_str)
                except ValueError:  # pragma: no cover
                    raise AssertionError('Switch {} is invalid. Use either PDB '
                                         'format or an int'.format(str(number_str)))

    def proc_num(self):
        """Return the number of the switch."""
        return self.sw_number

    @classmethod
    def parse_matrix_num(cls, num_str):
        """Parse a source/sink matrix tuple."""
        cr_list = num_str.split('/')
        return 32 + int(cr_list[0]) * 16 + int(cr_list[1])


class PDBCoil:

    """Base class for coils connected to a P-ROC/P3-ROC that are controlled via PDB driver boards.

    (i.e. the PD-16 board).
    """

    def __init__(self, pdb, number_str):
        """Find out number fo coil."""
        upper_str = number_str.upper()
        self.pdb = pdb
        if self.is_direct_coil(upper_str):
            self.coil_type = 'dedicated'
            self.banknum = (int(number_str[1:]) - 1) / 8
            self.outputnum = int(number_str[1:])
        elif self.is_pdb_coil(number_str):
            self.coil_type = 'pdb'
            (self.boardnum, self.banknum, self.outputnum) = pdb.decode_pdb_address(number_str)
        else:
            self.coil_type = 'unknown'

    def bank(self) -> int:
        """Return the bank number."""
        if self.coil_type == 'dedicated':
            return self.banknum
        if self.coil_type == 'pdb':
            return self.boardnum * 2 + self.banknum

        return -1

    def bank_secondary(self) -> Optional[int]:
        """Return the other bank if this is a PDB board."""
        if self.coil_type == 'pdb':
            if self.banknum == 0:
                return self.boardnum * 2 + 1

            return self.boardnum * 2

        return None

    def output(self):
        """Return the output number."""
        return self.outputnum

    @classmethod
    def is_direct_coil(cls, string):
        """Return true if it is a direct coil."""
        if len(string) < 2 or len(string) > 3:
            return False
        if not string[0] == 'C':
            return False
        if not string[1:].isdigit():
            return False
        return True

    def is_pdb_coil(self, string):
        """Return true if string looks like PDB address."""
        return self.pdb.is_pdb_address(string)


class PDBLight:

    """Base class for lights connected to a PD-8x8 driver board."""

    def __init__(self, pdb, number_str):
        """Find out light number."""
        self.pdb = pdb
        upper_str = number_str.upper()
        if self.is_direct_lamp(upper_str):
            self.lamp_type = 'dedicated'
            self.output = int(number_str[1:])
        elif self.is_pdb_lamp(number_str):
            # C-Ax-By-z:R-Ax-By-z  or  C-x/y/z:R-x/y/z
            self.lamp_type = 'pdb'
            source_addr, sink_addr = self.split_matrix_addr_parts(number_str)
            (self.source_boardnum, self.source_banknum, self.source_outputnum) = pdb.decode_pdb_address(source_addr)
            (self.sink_boardnum, self.sink_banknum, self.sink_outputnum) = pdb.decode_pdb_address(sink_addr)
        else:
            self.lamp_type = 'unknown'

    def source_board(self):
        """Return source board."""
        return self.source_boardnum

    def source_bank(self):
        """Return source bank."""
        return self.source_boardnum * 2 + self.source_banknum

    def sink_bank(self):
        """Return sink bank."""
        return self.sink_boardnum * 2 + self.sink_banknum

    def source_output(self):
        """Return source output."""
        return self.source_outputnum

    def sink_output(self):
        """Return sink output."""
        return self.sink_outputnum

    def dedicated_output(self):
        """Return dedicated output number."""
        return self.output

    @classmethod
    def is_direct_lamp(cls, string):
        """Return true if it looks like a direct lamp."""
        if len(string) < 2 or len(string) > 3:
            return False
        if not string[0] == 'L':
            return False
        if not string[1:].isdigit():
            return False
        return True

    @classmethod
    def split_matrix_addr_parts(cls, string) -> Tuple[str, str]:
        """Split the string of a matrix lamp address.

        Input is of form C-Ax-By-z:R-Ax-By-z  or  C-x/y/z:R-x/y/z  or
        aliasX:aliasY.  We want to return only the address part: Ax-By-z,
        x/y/z, or aliasX.  That is, remove the two character prefix if present.
        """
        addrs = string.rsplit(':')
        if len(addrs) != 2:
            raise AssertionError("Cannot split address {}. "
                                 "Expected form C-Ax-By-z:R-Ax-By-z or C-x/y/z:R-x/y/z".format(string))
        addrs_out = []
        for addr in addrs:
            bits = addr.split('-')
            if len(bits) == 1:
                addrs_out.append(addr)  # Append unchanged.
            else:  # Generally this will be len(bits) 2 or 4.
                # Remove the first bit and rejoin.
                addrs_out.append('-'.join(bits[1:]))
        assert len(addrs_out) == 2
        return addrs_out[0], addrs_out[1]

    def is_pdb_lamp(self, string):
        """Return true if it looks like a pdb lamp string."""
        params = self.split_matrix_addr_parts(string)
        if len(params) != 2:
            return False
        for addr in params:
            if not self.pdb.is_pdb_address(addr):
                return False
        return True


class PDBLED(PlatformBatchLight):

    """Represents an RGB LED connected to a PD-LED board."""

    __slots__ = ["board", "address", "debug", "log", "polarity", "platform"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, board, address, polarity, debug, driver_platform, light_system):
        """Initialise PDB LED."""
        self.board = board
        self.address = address
        self.debug = debug
        self.platform = driver_platform     # type: PROCBasePlatform
        super().__init__("{}-{}".format(self.board, self.address), light_system)
        self.log = logging.getLogger('PDBLED')
        self.polarity = polarity

        self.log.debug("Creating PD-LED item: board: %s, "
                       "RGB output: %s", self.board, self.address)

    def _normalise_color(self, value: int) -> int:
        if self.polarity:
            return 255 - value

        return value

    def get_max_fade_ms(self):
        """Return max fade ms."""
        return 4294967296

    def get_board_name(self):
        """Return board of the light."""
        return "PD-LED Board {}".format(self.board)

    def is_successor_of(self, other):
        """Return true if the other light has the previous address and is on the same board."""
        return self.board == other.board and self.address == other.address + 1

    def get_successor_number(self):
        """Return address on the same board."""
        return "{}-{}".format(self.board, self.address + 1)

    def __lt__(self, other):
        """Order lights by their position on the hardware."""
        return (self.board, self.address) < (other.board, other.address)


class PdLedServo(ServoPlatformInterface):

    """A servo on a PD-LED board."""

    def __init__(self, board, number, platform, debug):
        """Initialise PDB LED."""
        self.board = int(board)
        self.number = int(number)
        self.debug = debug
        self.log = logging.getLogger('PD-LED.Servo.{}-{}'.format(board, number))
        self.platform = platform    # type: PROCBasePlatform

    def go_to_position(self, position):
        """Move servo to a certain position."""
        value = int(position * 128) + 127
        if self.debug:
            self.log.debug("Setting servo to position: %s value: %s", position, value)

        self.platform.write_pdled_color(self.board, 72 + self.number, value)


class PdLedStepper(StepperPlatformInterface):

    """A stepper on a PD-LED board."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, board, number, platform, debug, stepper_ticks_per_half_period):
        """Initialise PDB LED."""
        self.board = int(board)
        self.number = int(number)
        self.debug = debug
        self.log = logging.getLogger('PD-LED.Stepper.{}-{}'.format(board, number))
        self.platform = platform
        self._move_complete = asyncio.Event()
        self._move_complete.set()
        self._move_timer = None
        self.stepper_ticks_per_half_period = stepper_ticks_per_half_period

    def move_vel_mode(self, velocity):
        """Turn stepper on at a certain speed."""
        if velocity == 0:
            self.stop()
        elif velocity > 0:
            self.move_rel_pos(16384)
        else:
            self.move_rel_pos(-16384)

    def move_rel_pos(self, position):
        """Move stepper by x steps."""
        if abs(position) > 16384:
            raise ValueError("Cannot move more than 16384 steps but tried {}".format(position))

        if position > 0:
            value = int(position)
        else:
            value = int(abs(position)) + (1 << 15)
        self.platform.write_pdled_config_reg(self.board, self.number + 23, value)

        self._move_complete.clear()
        # we need to time the steps and add 30ms for usb latency/jitter
        wait_time = ((int(abs(position)) * 2 * self.stepper_ticks_per_half_period) / 32000000) + 0.03
        self._move_timer = asyncio.sleep(wait_time)
        if self.debug:
            self.log.debug("Moving %s ticks. This will take %s", position, wait_time)
        self._move_timer = asyncio.ensure_future(self._move_timer)
        self._move_timer.add_done_callback(self._move_done)

    def _move_done(self, future):
        if self.debug:
            self.log.debug("Move done")
        try:
            future.result()
        except asyncio.CancelledError:
            return

        self._move_complete.set()

    def home(self, direction):
        """Not implemented."""
        del direction
        self.platform.raise_config_error("Use homing_mode switch for steppers on PD-LED.", 2)

    def stop(self):
        """Stop stepper."""
        self.platform.write_pdled_config_reg(self.board, self.number + 23, 0)
        self._move_complete.set()
        if self._move_timer:
            self._move_timer.cancel()

    async def wait_for_move_completed(self):
        """Wait for move complete."""
        return await self._move_complete.wait()



"""MMA8451 accelerometer platform."""
import asyncio
import logging

from mpf.core.utility_functions import Util

from mpf.platforms.interfaces.accelerometer_platform_interface import AccelerometerPlatformInterface

from mpf.core.platform import AccelerometerPlatform, I2cPlatform


class MMA8451Device(AccelerometerPlatformInterface):

    """MMA8451 accelerometer."""

    __slots__ = ["i2c_platform", "platform", "callback", "number", "task"]

    def __init__(self, number, callback, i2c_platform, platform):
        """Initialise MMA8451 accelerometer."""
        self.i2c_platform = i2c_platform    # type: I2cPlatform
        self.platform = platform            # type: MMA8451Platform
        self.callback = callback
        self.number = number

        self.task = self.platform.machine.clock.loop.create_task(self._poll())
        self.task.add_done_callback(Util.raise_exceptions)

    async def _poll(self):
        device = await self.i2c_platform.configure_i2c(self.number)
        # check id
        self.platform.log.info("Checking ID of device at: %s", self.number)
        device_id = await device.i2c_read8(0x0D)
        if device_id != 0x1A:
            raise AssertionError("Device ID does not match MMA8451. Detected: {}".format(device_id))

        # reset
        self.platform.log.info("Resetting device at: %s", self.number)
        device.i2c_write8(0x2B, 0x40)
        await asyncio.sleep(.3)
        result = -1
        for _ in range(10):
            result = await device.i2c_read8(0x2B)
            if result == 0:
                break
            self.platform.log.warning("Failed to reset: %s at %s", result, self.number)
            await asyncio.sleep(.5)
        else:
            raise AssertionError("Failed to reset MMA8451 accelerometer. Result: {}".format(result))

        # set resolution to 2g
        device.i2c_write8(0x2B, 0x02)

        # set ready
        device.i2c_write8(0x2D, 0x01)
        device.i2c_write8(0x2E, 0x01)

        # turn on orientation
        device.i2c_write8(0x11, 0x40)

        # low noise mode, 12,5Hz and activate
        device.i2c_write8(0x2A, 0x2D)

        # wait for activate
        await asyncio.sleep(.3)

        self.platform.log.info("Init done for device at: %s", self.number)

        while True:
            data = await device.i2c_read_block(0x01, 6)
            x = ((data[0] << 8) | data[1]) >> 2
            y = ((data[2] << 8) | data[3]) >> 2
            z = ((data[4] << 8) | data[5]) >> 2
            max_val = 2 ** (14 - 1) - 1
            signed_max = 2 ** 14
            x -= signed_max if x > max_val else 0
            y -= signed_max if y > max_val else 0
            z -= signed_max if z > max_val else 0
            range_divisor = 4096 / 9.80665
            x = round((float(x)) / range_divisor, 3)
            y = round((float(y)) / range_divisor, 3)
            z = round((float(z)) / range_divisor, 3)
            self.callback.update_acceleration(x, y, z)
            await asyncio.sleep(.1)


class MMA8451Platform(AccelerometerPlatform):

    """MMA8451 accelerometer platform."""

    __slots__ = ["accelerometers"]

    def __init__(self, machine):
        """Configure MMA8451 based accelerometers."""
        super().__init__(machine)
        self.log = logging.getLogger('mma8451')
        self.log.debug("Configuring MMA8451 based accelerometers.")
        self.accelerometers = {}

    async def initialize(self):
        """Initialise MMA8451 platform."""

    def stop(self):
        """Stop accelerometer poll tasks."""
        for accelerometer in self.accelerometers:
            if accelerometer.task:
                accelerometer.task.cancel()
                accelerometer.task = None
        self.accelerometers = {}

    def configure_accelerometer(self, number, config, callback) -> MMA8451Device:
        """Configure MMA8451 accelerometer."""
        config = self.machine.config_validator.validate_config("mma8451_accelerometer", config)
        i2c_platform = self.machine.get_platform_sections("i2c", config['i2c_platform'])
        i2c_platform.assert_has_feature("i2c")
        return MMA8451Device(number, callback, i2c_platform, self)



import asyncio
from typing import Tuple

from mpf.core.bcp.bcp_socket_client import decode_command_string, encode_command_string

from mpf.tests.loop import MockServer, MockQueueSocket

from mpf.tests.MpfTestCase import MpfTestCase


class TestVPX(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        del self.machine_config_patches['bcp']
        self.machine_config_patches['bcp'] = dict()
        self.machine_config_patches['bcp']['connections'] = []
        self.client = None

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/vpx/'

    def get_use_bcp(self):
        return True

    def get_platform(self):
        return False

    def _mock_loop(self):
        self.mock_server = MockServer(self.clock.loop)
        self.clock.mock_server("127.0.0.1", 5051, self.mock_server)

    def _early_machine_init(self, machine):
        org_init = machine._run_init_phases
        async def _init_hook():
            machine.events.add_async_handler("init_phase_4", self._send_init)
            await org_init()
        machine._run_init_phases = _init_hook

    async def _send_init(self, **kwargs):
        self.client = MockQueueSocket(self.loop)
        await self.mock_server.add_client(self.client)
        # check hello
        cmd, args = await self._get_and_decode(self.client)
        self.assertEqual("hello", cmd)

        self._encode_and_send("start")

    @staticmethod
    async def _get_and_decode(client) -> Tuple[str, dict]:
        data = await client.send_queue.get()
        return decode_command_string(data[0:-1].decode())

    def read_vpx_response_from_bcp(self):
        cmd, args = self.loop.run_until_complete(self._get_and_decode(self.client))
        self.assertEqual(cmd, "vpcom_bridge_response")
        self.assertNotIn("error", args, "Error: {}".format(args.get("error")))
        return args.get("result")

    def _encode_and_send(self, cmd, **kwargs):
        self.client.recv_queue.append((encode_command_string("vpcom_bridge", subcommand=cmd, **kwargs) + '\n').encode())

    def test_vpx(self):
        self.advance_time_and_run()
        self.client.send_queue = asyncio.Queue()

        self._encode_and_send("changed_lamps")
        self._encode_and_send("changed_solenoids")

        self.read_vpx_response_from_bcp()
        self.read_vpx_response_from_bcp()

        self.machine.lights["test_light1"].on()
        self.advance_time_and_run(.1)
        self._encode_and_send("changed_lamps")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['0', True]])

        self.machine.coils["c_test"].pulse()
        self.advance_time_and_run(.001)

        self._encode_and_send("changed_solenoids")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['2', True]])
        self.advance_time_and_run(.1)
        self._encode_and_send("changed_solenoids")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['2', False]])

        self.machine.coils["c_test"].enable()
        self.advance_time_and_run(.001)

        self._encode_and_send("changed_solenoids")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['2', True]])

        self.machine.coils["c_test"].disable()
        self.advance_time_and_run(.001)

        self._encode_and_send("changed_solenoids")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['2', False]])

        self.machine.flippers["f_test"].enable()
        self.advance_time_and_run(.001)
        self._encode_and_send("get_hardwarerules")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['3', '1', True]])

        self.machine.autofires["ac_slingshot_test"].enable()
        self.advance_time_and_run(.001)
        self._encode_and_send("get_hardwarerules")
        result = self.read_vpx_response_from_bcp()
        self.assertCountEqual(result, [['3', '1', True], ['0', '0', False]])

        self.machine.flippers["f_test"].disable()
        self.advance_time_and_run(.001)
        self._encode_and_send("get_hardwarerules")
        result = self.read_vpx_response_from_bcp()
        self.assertEqual(result, [['0', '0', False]])

        self.assertSwitchState("s_test", False)
        self._encode_and_send("set_switch", number=6, value=1)
        self.advance_time_and_run(.1)
        self.read_vpx_response_from_bcp()
        self.assertSwitchState("s_test", True)
        self._encode_and_send("set_switch", number=6, value=0)
        self.advance_time_and_run(.1)
        self.read_vpx_response_from_bcp()
        self.assertSwitchState("s_test", False)



from mpf.plugins.auditor import Auditor
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestAuditor(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_platform(self):
        return 'smart_virtual'

    def get_machine_path(self):
        return 'tests/machine_files/auditor/'

    def setUp(self):
        self.machine_config_patches['mpf']['plugins'] = ['mpf.plugins.auditor.Auditor']
        super().setUp()

    def test_auditor_player_vars(self):
        auditor = self.machine.plugins[0]
        self.assertIsInstance(auditor, Auditor)
        data_manager = auditor.data_manager

        # start a game
        self.start_game()

        self.post_event("add_score")
        self.post_event("add_custom")
        self.post_event("add_custom")
        self.post_event("add_not_audited")
        self.assertPlayerVarEqual(100, "score")

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.assertEqual(100, auditor.current_audits['player']['score']['average'])
        self.assertEqual([100], auditor.current_audits['player']['score']['top'])
        self.assertEqual(1, auditor.current_audits['player']['score']['total'])
        self.assertEqual(200, auditor.current_audits['player']['my_var']['average'])
        self.assertEqual([200], auditor.current_audits['player']['my_var']['top'])
        self.assertEqual(1, auditor.current_audits['player']['my_var']['total'])
        self.assertNotIn("not_audited", auditor.current_audits['player'])

        # start a game
        self.start_game()

        self.post_event("add_score")
        self.post_event("add_score")
        self.assertPlayerVarEqual(200, "score")

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.assertEqual(150, auditor.current_audits['player']['score']['average'])
        self.assertEqual([200, 100], auditor.current_audits['player']['score']['top'])
        self.assertEqual(2, auditor.current_audits['player']['score']['total'])
        self.assertEqual(100, auditor.current_audits['player']['my_var']['average'])
        self.assertEqual([200, 0], auditor.current_audits['player']['my_var']['top'])
        self.assertEqual(2, auditor.current_audits['player']['my_var']['total'])
        self.assertNotIn("not_audited", auditor.current_audits['player'])

        self.assertEqual({'score': {'top': [200, 100], 'average': 150.0, 'total': 2},
                          'my_var': {'top': [200, 0], 'average': 100.0, 'total': 2}},
                         auditor.data_manager.written_data["player"])

    def test_auditor_switches(self):
        auditor = self.machine.plugins[0]
        self.assertIsInstance(auditor, Auditor)
        data_manager = auditor.data_manager
        self.machine.switch_controller.process_switch("s_ball", 1)

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(2)

        self.assertEqual(0, auditor.current_audits['switches']['s_test'])
        self.assertMachineVarEqual(0, "audits_switches_s_test")

        # start a game
        self.start_game()

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(1)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.assertEqual(1, auditor.current_audits['switches']['s_test'])
        self.assertMachineVarEqual(1, "audits_switches_s_test")

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(.1)

        self.assertEqual(1, auditor.current_audits['switches']['s_test'])
        self.assertMachineVarEqual(1, "audits_switches_s_test")
        self.assertEqual(1, data_manager.written_data['switches']['s_test'])

        # start a game
        self.start_game()

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(.1)

        self.assertEqual(2, auditor.current_audits['switches']['s_test'])
        self.assertMachineVarEqual(2, "audits_switches_s_test")
        self.assertEqual(2, data_manager.written_data['switches']['s_test'])

        self.post_event("test_event1")
        self.post_event("test_event2")
        self.post_event("test_event3")
        self.advance_time_and_run(.1)

        self.assertMachineVarEqual(2, "audits_switches_s_test")
        self.assertEqual(1, data_manager.written_data['events']['test_event1'])
        self.assertEqual(1, data_manager.written_data['events']['test_event2'])
        self.assertNotIn("test_event3", data_manager.written_data['events'])

        self.post_event("test_event1")
        self.advance_time_and_run(.1)

        self.assertEqual(2, data_manager.written_data['events']['test_event1'])
        self.assertEqual(1, data_manager.written_data['events']['test_event2'])

        # should not crash on unknown switch
        self.machine.switch_controller.process_switch_by_num(123123123123, 1, self.machine.default_platform)
        self.advance_time_and_run(.1)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(.1)

        self.assertEqual(2, auditor.current_audits['switches']['s_test'])
        self.assertEqual(2, data_manager.written_data['switches']['s_test'])

        self.post_event("auditor_reset")
        self.advance_time_and_run(.1)

        self.assertEqual(0, auditor.current_audits['switches']['s_test'])
        self.assertEqual(0, data_manager.written_data['switches']['s_test'])
